////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

// Last update: 16 May 1996 (improved table AI)

using namespace std;

#include <iostream>
#include <iomanip>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <complex>
#include "common.hpp"
#include "dyna.hpp"
#include "timer.hpp"
#include "info.hpp"
#include "element.hpp"
#include "vector.hpp"
#include "store.hpp"
#include "avl.hpp"
#include "sparse-nocomplex.hpp"
#include "hydra-steady.hpp"

// ### work out exact core needing replacement for TABLE_AI

#define TABLE_AI

void SteadyStateAnalyser::aiPrint(double a, double b, double c, double d, double e, double f)
{ 
  cout << "aiPrint: " << endl;
  cout << setw(12) << 1-a-c << " " << setw(12) << c << " " << setw(12) << a << endl;
  cout << setw(12) << d << " " << setw(12) << 1-b-d << " " << setw(12) << b << endl;
  cout << setw(12) << f << " " << setw(12) << e << " " << setw(12) << 1-e-f << endl;
}

#ifndef TABLE_AI
void SteadyStateAnalyser::air(Vector<double> &solution, long maxIterations, double accuracy, int refine)
{
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting AIR solve..." << endl;
  #endif

  register double omega = 1.0;
  register double aa, cc, ee, ff, a, b, c, d, e, f, sum;
  register long i, k;
  register int rk;
  register SparseMatrixCol *_col;
  register double l, pi, u, _u; // , ll, uu;
  double maxDiff, solNorm, lastResidInfNorm, residInfNorm;
  double improve, convergence = 1.0;
  long iterations = 0;
  Vector<double> last, result;
#ifdef EXPERIMENT
  Vector<double> normSol, normRes;
  normSol.setLength(tangible);
  normRes.setLength(tangible);
#endif
  double tol = (double) EPS/tangible;

  int count[41], index = 0, maxIndex = 0, stop = 0;
  double average[41], max = 0, step = 0.05;

  for (i=0; i<41; i++) {
    count[i] = 0;
    average[i] = 0;
  }

  #ifdef DEBUG
  QC->print();
  #endif

  last.setLength(tangible);
  result.setLength(tangible);

  if (!refine) timer->read();

  double QC01 = QC->getElement(0, 1),
       QCnn = QC->getElement(tangible - 1, tangible - 1),
       QCnn_1 = QC->getElement(tangible - 1, tangible - 2);

  // initialise solution, last vectors
  for (i=0; i<tangible; i++) {
    last[i] = solution(i);
//    cout << "solution(" << i << ") = " << solution[i] << endl;
  }

  /* AIR extension */
  Q->transMultiply(solution,result);
  for (i=0; i<tangible; i++)
    result[i] -= solution[i];
  lastResidInfNorm = result.infNorm();
  #ifdef EXPERIMENT
    cout << "iteration " << 0 << " infNorm=" << result.infNorm() << " twoNorm=" << result.twoNorm() << endl;
    cout << "EXPERIMENT " << 0 << " " << log10(result.twoNorm()) << endl;  
  #endif

  if (solRepStyle != NONE) {
    cout << "   Iteration       Omega                Resid Inf. Norm      Accuracy" << endl;
    cout << "----------------+--------------------+--------------------+----------------" << endl;
  }
  cout.setf(ios::showpoint);

// ###
  while (iterations < maxIterations && convergence > accuracy) {
    l = solution(0);
    u = 1.0 - (solution(0) + solution(1));

//////////////////////////////////////////////////////////////////////

 /* aiCalcFirstParameters(1, a,ai_b,c,ai_d,e,f, solution, l, u);*/

// ###
//    if (l < tol) l = tol;

    c = QC01;

    aa = ai_start*l;
//    assert(fabs(ai_start-u)<EPS);
    //    cout << "aa start = " << ai_start << " u = " << u << endl;
    a = ai_start;

    _col = &QC->getCol(0);
    ff = 0;
    for (rk = 0; rk<_col->getEntries(); rk++) {
      k = _col->getRow(rk);
      if (k > 1) ff += solution(k)*QC->getColElement(_col, rk);
    }
    f = ff / u;

    _col = &QC->getCol(1);
    ee = 0;
    for (rk=0; rk<_col->getEntries(); rk++) {
      k = _col->getRow(rk);
      if (k > 1) ee += solution(k) * QC->getColElement(_col, rk);
    }
    e = ee/u;

//////////////////////////////////////////////////////////////////////
/*    aiSolve(a,ai_b(1),c,ai_d(1),e,f,l,pi,u); */
    b = ai_b(1);
    d = ai_d(1);
//    cout << "FIRST MATRIX:" << endl;
//    aiPrint(a,b,c,d,e,f);   
//   if (a > 1) cout << "BLUE MURDER! A > 1 ON FIRST ITERATION" << endl;
// ### we are changing here
    sum = 1/(e + f);
    a *= sum;
    b *= sum;
    c += a*e;
    d += b*f;
    
    pi = c/d;

    _u = a + pi*b;

    sum = (double) 1.0/(1.0 + pi + _u);
    l = sum;
    pi *= sum;

    if (l < tol) l = tol;
    if (pi < tol) pi = tol;

//    cout << "1) l = " << l << " pi = " << pi << " u = " << _u*sum << endl;

//////////////////////////////////////////////////////////////////////
    solution[0] = l;
    solution[1] = pi;

//    cout << "solution(0) = " << solution(0) << endl;
//    cout << "solution(1) = " << solution(1) << endl;
    l = solution(0) + solution(1);
//  l += solution(1);
    for (i = 2; i < tangible - 2; i++) {
      _col = &QC->getCol(i);
      d = ai_d(i);
//        u = 1 - l - solution(i);
      u -= solution(i);
//      u = 0;
//      for (k = i; k < tangible; k++) {
//        u += solution(k);
//      }

//////////////////////////////////////////////////////////////////////
 /*      aiCalcParameters(n, a,ai_b,c,ai_d,e,f, solution, l, u); */

// ###
//      if (l < tol) {
//        cout << "warning!!! (" << i << " iteration, l = " << l << ") " << endl;
// 	l = tol;
//      }
// ### added
//      ll = 1/l; uu = 1/u;

      ff = ff + ee - solution(i)*d; //ai_d(i) + ee;
      f = ff/u;
     
      // c is from (L) to i
      // e is from (R) to i
      cc = 0; ee = 0;
      for (rk=0; rk<_col->getEntries(); rk++) {
	k = _col->getRow(rk);
	if (k < i)
          cc += solution(k) * QC->getColElement(_col, rk);
        else if (k > i)
          ee += solution(k) * QC->getColElement(_col, rk);
      }
      c = cc/l; e = ee/u;

      // a is from L to R
      aa = aa - cc + solution(i-1) * ai_b(i-1); // - cc;
      a = aa/l;      
//      if (a > 1.0) cout << "WARNING: a = " << a << " (i= " << i << ", l=" << l << ")" << endl;

//////////////////////////////////////////////////////////////////////
    /* aiSolve(a,ai_b(i),c,ai_d(i),e,f,l,pi,u); */

      sum = 1/(e+f);      
      b = ai_b(i);
//      d = ai_d(i);

//      aiPrint(a,b,c,d,e,f);
// ### we are changing here
      a *= sum;
      b *= sum;
      c += a*e;
      d += b*f;

      pi = c/d;

      _u = a + pi*b;
      
      sum = (double) 1.0/(1.0 + pi + _u);
      l = sum; // ### tres questionable
      pi *= sum;
      if (pi < tol) pi = tol;

//      cout << "i) l = " << l << " pi = " << pi << " u = " << _u*sum << endl;

//////////////////////////////////////////////////////////////////////
      solution[i] = pi; // *omega + (1-omega)*last(i);
   
      l += solution(i);
    }

    u -= solution(tangible - 2);
//////////////////////////////////////////////////////////////////////

 /* aiCalcLastParameters(tangible - 2, a,ai_b,c,ai_d,e,f, solution, l, u);*/
// ###
//    if (l < tol) l = tol;

    i = tangible - 2;
    d = ai_d(i); // ###
    _col = &QC->getCol(i);

    cc = 0;
    for (rk=0; rk<_col->getEntries(); rk++) {
      k = _col->getRow(rk);
      if (k < i) cc += solution(k)*QC->getColElement(_col, rk);
    }
    c = cc/l;
    
    aa = aa + solution(i-1)*ai_b(i-1) - cc;
    a = aa/l;

    e = QCnn_1;
    f = 1 - (QCnn + e);

//////////////////////////////////////////////////////////////////////

/*  aiSolve(a,ai_b(tangible - 2),c,ai_d(tangible - 2),e,f,l,pi,u); */

    b = ai_b(i);
//    d = ai_d(i);

//    aiPrint(a,b,c,d,e,f);
// ### we are changing here
    sum = 1/(e+f);
    a *= sum;
    b *= sum;
    c += a*e;
    d += b*f;

    pi = c/d;
    u = a + pi*b;

    sum = (double) 1.0/(1.0 + pi + u);
    l = sum;
    pi *= sum;
    u *= sum;

    if (pi < tol) pi = tol;
    if (u < tol) u = tol;

//    cout << "n) l = " << l << " pi = " << pi << " u = " << u << endl;

//////////////////////////////////////////////////////////////////////

    solution[tangible - 2] = pi;
    solution[tangible - 1] = u;

//    cout << "solution(" << tangible - 2 << ") = " << pi << endl;
//    cout << "solution(" << tangible - 1 << ") = " << u << endl;
//    cout << "solution's one norm is " << solution.oneNorm() << endl;
    solution.normalise();

    for (i=0; i<tangible; i++) {
      solution[i] = (1-omega)*last(i) + omega*solution(i);
      if (solution[i] < tol) solution[i] = tol;
    }
    solution.normalise();

#ifdef DEBUG
    cout << "solution: " << solution << endl;
#endif
    iterations++;

#ifdef EXPERIMENT
    for (i=0; i<tangible; i++)
      normSol[i] = solution[i];
    normSol.normalise();
    Q->transMultiply(normSol, normRes);
    for (i=0; i<tangible; i++)
      normRes[i] -= normSol[i];
    cout << "iteration " << iterations << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << iterations*1.2 << " " << log10(normRes.twoNorm()) << endl;
#endif

    if (!(iterations % 5)) {
      maxDiff = 0;
      solNorm = 0;
      for (i = 0; i < tangible; i++) {
        if (fabs(solution(i)) > solNorm) solNorm = fabs(solution(i));
        if (fabs(solution(i)-last(i)) > maxDiff)
          maxDiff = fabs(solution(i) - last(i));
        last(i) = solution(i);
      }
      if (maxDiff != 0) convergence = maxDiff/solNorm;
        else convergence = 0;

      residInfNorm = 0;
      QC->transMultiply(solution, result);
      for (i = 0; i<tangible; i++) {
	result[i] -= solution(i);
	if (fabs(result(i)) > residInfNorm)
	  residInfNorm = fabs(result(i));
      }
      improve = (lastResidInfNorm-residInfNorm)/lastResidInfNorm;
//      cout << "impr = " << improve << " res = " << residInfNorm << endl;
 //      i = (long) ((omega-1)*40+0.5);
//      cout << "omega =  " << omega << " index = " << index << endl;
      if (iterations == 5) improve = 1.0;
      if (improve < 0) {
      omega = 1+(omega - 1)/2;
      step = step * 0.5;
      if (step < 0.005) step = 0.005;
    } else {
      omega += step;
    }

//      cout << "choose " << omega << " (" << omega << ") with max = " << max << endl;
      lastResidInfNorm = residInfNorm;

      if (solRepStyle != NONE) {
	cout << setw(12) << iterations << "    | " << setw(15) << omega << "    | " << setw(15) << residInfNorm << "    | " << setw(15) << convergence << endl;
      }
    }
  };

  if (convergence > accuracy)
    terminate("AIR failed to converge within %d iterations", maxIterations);

  residInfNorm = 0;
  QC->transMultiply(solution, result);
  for (i = 0; i<tangible; i++) {
    result[i] -= solution(i);
    if (fabs(result(i)) > residInfNorm)
      residInfNorm = fabs(result(i));
  }

  if (solRepStyle != NONE)
    cout << "----------------+--------------------+--------------------+----------------" << endl;

  cout.unsetf(ios::showpoint);
  checkViolation(checkVector(solution, residInfNorm));
  if (refine) {
    cout << "SteadyStateAnalyser (" << this << "): AIR refine complete." << endl;
    return;
  }
  unScale(solution);
  solution.normalise();
  double timeUsed = timer->read();
  solution.output(steadyVectorFile); 
  #ifndef RELEASE
  solution.textOutput("air.sol");
  #endif
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): AIR complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}
#else
void SteadyStateAnalyser::air(Vector<double> &solution, long maxIterations, double accuracy, int refine)
{
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting AIR solve..." << endl;
  #endif

  register double omega = 1.0;
  register double aa, cc, ee, ff, a, b, c, d, e, f, sum;
  register long i, k;
  register int rk;
  register SparseMatrixCol *_col;
  register double l, pi, u, _u; // , ll, uu;
  double maxDiff, solNorm, lastResidInfNorm, residInfNorm;
  double improve, convergence = 1.0;
  long iterations = 0;
  Vector<double> last, result;
#ifdef EXPERIMENT
  Vector<double> normSol, normRes;
  normSol.setLength(tangible);
  normRes.setLength(tangible);
#endif
  double tol = (double) EPS/tangible;

  int count[41], index = 0, maxIndex = 0, stop = 0;
  double average[41], max = 0;

  for (i=0; i<41; i++) {
    count[i] = 0;
    average[i] = 0;
  }

  #ifdef DEBUG
  QC->print();
  #endif

  last.setLength(tangible);
  result.setLength(tangible);

  if (!refine) timer->read();

  double QC01 = QC->getElement(0, 1),
       QCnn = QC->getElement(tangible - 1, tangible - 1),
       QCnn_1 = QC->getElement(tangible - 1, tangible - 2);

  // initialise solution, last vectors
  for (i=0; i<tangible; i++) {
    last[i] = solution(i);
//    cout << "solution(" << i << ") = " << solution[i] << endl;
  }

  // AIR extension 
  Q->transMultiply(solution,result);
  for (i=0; i<tangible; i++)
    result[i] -= solution[i];
  lastResidInfNorm = result.infNorm();
  #ifdef EXPERIMENT
    cout << "iteration " << 0 << " infNorm=" << result.infNorm() << " twoNorm=" << result.twoNorm() << endl;
    cout << "EXPERIMENT " << 0 << " " << log10(result.twoNorm()) << endl;  
  #endif

  if (solRepStyle != NONE) {
    cout << "   Iteration       Omega                Resid Inf. Norm      Accuracy" << endl;
    cout << "----------------+--------------------+--------------------+----------------" << endl;
  }
  cout.setf(ios::showpoint);

// ###
  while (iterations < maxIterations && convergence > accuracy) {
    l = solution(0);
    u = 1.0 - (solution(0) + solution(1));

//////////////////////////////////////////////////////////////////////

 // aiCalcFirstParameters(1, a,ai_b,c,ai_d,e,f, solution, l, u);

// ###
//    if (l < tol) l = tol;

    c = QC01;

    aa = ai_start*l;
//    assert(fabs(ai_start-u)<EPS);
//    cout << "aa start = " << ai_start << " u = " << u << endl;
    a = ai_start;

    _col = &QC->getCol(0);
    ff = 0;
    for (rk = 0; rk<_col->getEntries(); rk++) {
      k = _col->getRow(rk);
      if (k > 1) ff += solution(k)*QC->getColElement(_col, rk);
    }
    f = ff / u;

    _col = &QC->getCol(1);
    ee = 0;
    for (rk=0; rk<_col->getEntries(); rk++) {
      k = _col->getRow(rk);
      if (k > 1) ee += solution(k) * QC->getColElement(_col, rk);
    }
    e = ee/u;

//////////////////////////////////////////////////////////////////////
 //    aiSolve(a,ai_b(1),c,ai_d(1),e,f,l,pi,u); 
    b = ai_b(1);
    d = ai_d(1);
//    cout << "FIRST MATRIX:" << endl;
//    aiPrint(a,b,c,d,e,f);   
//   if (a > 1) cout << "BLUE MURDER! A > 1 ON FIRST ITERATION" << endl;
// ### we are changing here
    sum = 1/(e + f);
    a *= sum;
    b *= sum;
    c += a*e;
    d += b*f;
    
    pi = c/d;

    _u = a + pi*b;

    sum = (double) 1.0/(1.0 + pi + _u);
    l = sum;
    pi *= sum;

    if (l < tol) l = tol;
    if (pi < tol) pi = tol;

//    cout << "1) l = " << l << " pi = " << pi << " u = " << _u*sum << endl;

//////////////////////////////////////////////////////////////////////
    solution[0] = l;
    solution[1] = pi;

//    cout << "solution(0) = " << solution(0) << endl;
//    cout << "solution(1) = " << solution(1) << endl;
    l = solution(0) + solution(1);
//  l += solution(1);
    for (i = 2; i < tangible - 2; i++) {
      _col = &QC->getCol(i);
      d = ai_d(i);
//        u = 1 - l - solution(i);
      u -= solution(i);
//      u = 0;
//      for (k = i; k < tangible; k++) {
//        u += solution(k);
//      }

//////////////////////////////////////////////////////////////////////
 //      aiCalcParameters(n, a,ai_b,c,ai_d,e,f, solution, l, u); 

// ###
//      if (l < tol) {
//        cout << "warning!!! (" << i << " iteration, l = " << l << ") " << endl;
// 	l = tol;
//      }
// ### added
//      ll = 1/l; uu = 1/u;

      ff = ff + ee - solution(i)*d; //ai_d(i) + ee;
      f = ff/u;
     
      // c is from (L) to i
      // e is from (R) to i
      cc = 0; ee = 0;
      for (rk=0; rk<_col->getEntries(); rk++) {
	k = _col->getRow(rk);
	if (k < i)
          cc += solution(k) * QC->getColElement(_col, rk);
        else if (k > i)
          ee += solution(k) * QC->getColElement(_col, rk);
      }
      c = cc/l; e = ee/u;

      // a is from L to R
      aa = aa - cc + solution(i-1) * ai_b(i-1); // - cc;
      a = aa/l;      
//      if (a > 1.0) cout << "WARNING: a = " << a << " (i= " << i << ", l=" << l << ")" << endl;

//////////////////////////////////////////////////////////////////////
    // aiSolve(a,ai_b(i),c,ai_d(i),e,f,l,pi,u); 

      sum = 1/(e+f);      
      b = ai_b(i);
//      d = ai_d(i);

//      aiPrint(a,b,c,d,e,f);
// ### we are changing here
      a *= sum;
      b *= sum;
      c += a*e;
      d += b*f;

      pi = c/d;

      _u = a + pi*b;
      
      sum = (double) 1.0/(1.0 + pi + _u);
      l = sum; // ### tres questionable
      pi *= sum;
      if (pi < tol) pi = tol;

//      cout << "i) l = " << l << " pi = " << pi << " u = " << _u*sum << endl;

//////////////////////////////////////////////////////////////////////
      solution[i] = pi; // *omega + (1-omega)*last(i);
   
      l += solution(i);
    }

    u -= solution(tangible - 2);
//////////////////////////////////////////////////////////////////////

 // aiCalcLastParameters(tangible - 2, a,ai_b,c,ai_d,e,f, solution, l, u);
// ###
//    if (l < tol) l = tol;

    i = tangible - 2;
    d = ai_d(i); // ###
    _col = &QC->getCol(i);

    cc = 0;
    for (rk=0; rk<_col->getEntries(); rk++) {
      k = _col->getRow(rk);
      if (k < i) cc += solution(k)*QC->getColElement(_col, rk);
    }
    c = cc/l;
    
    aa = aa + solution(i-1)*ai_b(i-1) - cc;
    a = aa/l;

    e = QCnn_1;
    f = 1 - (QCnn + e);

//////////////////////////////////////////////////////////////////////

//  aiSolve(a,ai_b(tangible - 2),c,ai_d(tangible - 2),e,f,l,pi,u); 

    b = ai_b(i);
//    d = ai_d(i);

//    aiPrint(a,b,c,d,e,f);
// ### we are changing here
    sum = 1/(e+f);
    a *= sum;
    b *= sum;
    c += a*e;
    d += b*f;

    pi = c/d;
    u = a + pi*b;

    sum = (double) 1.0/(1.0 + pi + u);
    l = sum;
    pi *= sum;
    u *= sum;

    if (pi < tol) pi = tol;
    if (u < tol) u = tol;

//    cout << "n) l = " << l << " pi = " << pi << " u = " << u << endl;

//////////////////////////////////////////////////////////////////////

    solution[tangible - 2] = pi;
    solution[tangible - 1] = u;

//    cout << "solution(" << tangible - 2 << ") = " << pi << endl;
//    cout << "solution(" << tangible - 1 << ") = " << u << endl;
//    cout << "solution's one norm is " << solution.oneNorm() << endl;
    solution.normalise();

    for (i=0; i<tangible; i++) {
      solution[i] = (1-omega)*last(i) + omega*solution(i);
      if (solution[i] < tol) solution[i] = tol;
    }
    solution.normalise();

#ifdef DEBUG
    cout << "solution: " << solution << endl;
#endif
    iterations++;

#ifdef EXPERIMENT
    for (i=0; i<tangible; i++)
      normSol[i] = solution[i];
    normSol.normalise();
    Q->transMultiply(normSol, normRes);
    for (i=0; i<tangible; i++)
      normRes[i] -= normSol[i];
    cout << "iteration " << iterations << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << iterations*1.2 << " " << log10(normRes.twoNorm()) << endl;
#endif

    if (!(iterations % 5)) {
      maxDiff = 0;
      solNorm = 0;
      for (i = 0; i < tangible; i++) {
        if (fabs(solution(i)) > solNorm) solNorm = fabs(solution(i));
        if (fabs(solution(i)-last(i)) > maxDiff)
          maxDiff = fabs(solution(i) - last(i));
        last(i) = solution(i);
      }
      if (maxDiff != 0) convergence = maxDiff/solNorm;
        else convergence = 0;

      residInfNorm = 0;
      QC->transMultiply(solution, result);
      for (i = 0; i<tangible; i++) {
	result[i] -= solution(i);
	if (fabs(result(i)) > residInfNorm)
	  residInfNorm = fabs(result(i));
      }
      improve = (lastResidInfNorm-residInfNorm)/lastResidInfNorm;
      if (improve < 0) improve = 0;
//      cout << "impr = " << improve << " res = " << residInfNorm << endl;
 //      i = (long) ((omega-1)*40+0.5);
//      cout << "omega =  " << omega << " index = " << index << endl;
      if (!count[index])
	average[index] = improve;
      else {
	if (improve > average[index])
	  average[index] = 0.1*average[index] + 0.9*improve;
	else 
	  average[index] = 0.3*average[index] + 0.7*improve;
      }
      if (average[index] > max) {
	max = average[index];
	maxIndex = index;
      } else {
	max = 0;
	for (i=0; i<41; i++)
	  if (average[i] > max) {
	    max = average[i];
	    maxIndex = i;
	  }
      }
//      cout << "average[" << index << "] = " << average[index] << endl;
      count[index]++;
      if (improve <= 0 && iterations > 5) stop = 1;
      if (iterations < 200 && !stop) {
	omega += 0.05;
	index += 2;
      } else {
	omega = (0.025*maxIndex) + 1.0;
	index = maxIndex;
        if (index<40 && count[index+1]==0) {
   	  omega += 0.025;
	  index++;
        }
      }
      if (omega > 2.0) {
	omega = 2.0;
	index = 40;
      }
//      cout << "choose " << omega << " (" << omega << ") with max = " << max << endl;
      lastResidInfNorm = residInfNorm;

      if (solRepStyle != NONE) {
	cout << setw(12) << iterations << "    | " << setw(15) << omega << "    | " << setw(15) << residInfNorm << "    | " << setw(15) << convergence << endl;
      }
    }
  };

  if (convergence > accuracy)
    terminate("AIR failed to converge within %d iterations", maxIterations);

  residInfNorm = 0;
  QC->transMultiply(solution, result);
  for (i = 0; i<tangible; i++) {
    result[i] -= solution(i);
    if (fabs(result(i)) > residInfNorm)
      residInfNorm = fabs(result(i));
  }

  if (solRepStyle != NONE)
    cout << "----------------+--------------------+--------------------+----------------" << endl;

  cout.unsetf(ios::showpoint);
  checkViolation(checkVector(solution, residInfNorm));
  if (refine) {
    cout << "SteadyStateAnalyser (" << this << "): AIR refine complete." << endl;
    return;
  }
  unScale(solution);
  solution.normalise();
  double timeUsed = timer->read();
  solution.output(steadyVectorFile); 
  #ifndef RELEASE
  solution.textOutput("air.sol");
  #endif
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): AIR complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}
#endif

void SteadyStateAnalyser::ai(Vector<double> &solution, long maxIterations, double accuracy, int refine)
{
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): starting AI solve..." << endl;
  #endif
  register double aa, cc, ee, ff, a, b, c, d, e, f, sum;
  register long i, k;
  register int rk;
  register SparseMatrixCol *_col;
  register double l, pi, u, _u;
  double maxDiff, solNorm, residInfNorm, convergence = 1.0;
  long iterations = 0;
  Vector<double> last, result;
  // ### PUT ME BACK!
  double tol = (double) EPS/tangible; ///tangible;
#ifdef EXPERIMENT
  Vector<double> normSol, normRes;
  normSol.setLength(tangible);
  normRes.setLength(tangible);
#endif

  #ifdef DEBUG
  QC->print();
  #endif

  last.setLength(tangible);
  result.setLength(tangible);

  if (!refine) timer->read();

  double QC01 = QC->getElement(0, 1),
       QCnn = QC->getElement(tangible - 1, tangible - 1),
       QCnn_1 = QC->getElement(tangible - 1, tangible - 2);

  // initialise solution, last vectors
  for (i=0; i<tangible; i++) {
    last[i] = solution(i);
//    cout << "solution(" << i << ") = " << solution[i] << endl;
  }

  if (solRepStyle != NONE) {
    cout << "   Iteration                          Accuracy" << endl;
    cout << "----------------------------------+---------------------------------------" << endl;
  }
  cout.setf(ios::showpoint);

#ifdef EXPERIMENT
  Q->transMultiply(solution,result);
  for (i=0; i<tangible; i++)
    result[i] -= solution[i];
    cout << "iteration " << 0 << " infNorm=" << result.infNorm() << " twoNorm=" << result.twoNorm() << endl;
    cout << "EXPERIMENT " << 0 << " " << log10(result.twoNorm()) << endl;  
#endif

// ###
  while (iterations < maxIterations && convergence > accuracy) {
    l = solution(0);
    u = 1.0 - (solution(0) + solution(1));

//////////////////////////////////////////////////////////////////////

 // aiCalcFirstParameters(1, a,ai_b,c,ai_d,e,f, solution, l, u);

// ###
//    if (l < tol) l = tol;

    c = QC01;

   // cout << "c start = " << c << endl;

    aa = ai_start*l;
//    assert(fabs(ai_start-u)<EPS);
    //cout << "aa start = " << aa << " l = " << l << " u = " << u << endl;
    a = ai_start;
    //cout << "a start = " << a << endl;

    _col = &QC->getCol(0);
    ff = 0;
    for (rk = 0; rk<_col->getEntries(); rk++) {
      k = _col->getRow(rk);
      if (k > 1) ff += solution(k)*QC->getColElement(_col, rk);
    }
    //cout << "ff start = " << ff << endl;
    f = ff / u;
    //cout << "f start = " << f << endl;

    _col = &QC->getCol(1);
    ee = 0;
    for (rk=0; rk<_col->getEntries(); rk++) {
      k = _col->getRow(rk);
      if (k > 1) ee += solution(k) * QC->getColElement(_col, rk);
    }
    e = ee/u;

    //cout << "ee start = " << ee << endl;
    //cout << "e start = " << e << endl;

//////////////////////////////////////////////////////////////////////
//    aiSolve(a,ai_b(1),c,ai_d(1),e,f,l,pi,u); 
    b = ai_b(1);
    d = ai_d(1);
    
    //    cout << "FIRST MATRIX:" << endl;
    // aiPrint(a,b,c,d,e,f);   

    if (a > 1) cout << "BLUE MURDER! A > 1 ON FIRST ITERATION" << endl;
    sum = (e + f);
    a /= sum;
    b /= sum;
    c += a*e;
    d += b*f;
    
    pi = c/d;

    _u = a + pi*b;

    sum = (double) 1.0/(1.0 + pi + _u);
    l = sum;
    pi *= sum;

    if (l < tol) l = tol;
    if (pi < tol) pi = tol;

//    cout << "1) l = " << l << " pi = " << pi << " u = " << _u*sum << endl;

//////////////////////////////////////////////////////////////////////
    solution[0] = l;
    solution[1] = pi;

    //cout << "solution(0) = " << solution(0) << endl;
    //cout << "solution(1) = " << solution(1) << endl;
    l = solution(0) + solution(1);
//  l += solution(1);
    for (i = 2; i < tangible - 2; i++) {
      _col = &QC->getCol(i);
      d = ai_d(i);
//        u = 1 - l - solution(i);
      u -= solution(i);

      // ### remove this
      // double _l = 0.0;
      /*      for (k=0; k<i; k++) {
	_l += solution(k);
      }
      cout << "l = " << l << " _l = " << _l << endl;
      assert(fabs(l - _l) < 0.1)*/


//      u = 0;
//      for (k = i; k < tangible; k++) {
//        u += solution(k);
//      }

//////////////////////////////////////////////////////////////////////
 //      aiCalcParameters(n, a,ai_b,c,ai_d,e,f, solution, l, u); 

// ###
      //      if (l < tol) l = tol
//      if (l < tol) {
//        cout << "warning!!! (" << i << " iteration, l = " << l << ") " << endl;
// 	l = tol;
//      }
      ff = ff + ee - solution(i)*d; //ai_d(i) + ee;
      f = ff/u;
     
      // c is from (L) to i
      // e is from (R) to i
      cc = 0; ee = 0;
      for (rk=0; rk<_col->getEntries(); rk++) {
	k = _col->getRow(rk);
	if (k < i)
          cc += solution(k) * QC->getColElement(_col, rk);
        else if (k > i)
          ee += solution(k) * QC->getColElement(_col, rk);
      }
      c = cc/l; e = ee/u;
      
      // a is from L to R
      aa = aa - cc + solution(i-1) * ai_b(i-1); // - cc;
      a = aa/l;      
//      if (a > 1.0) cout << "WARNING: a = " << a << " (i= " << i << ", l=" << l << ")" << endl;

//////////////////////////////////////////////////////////////////////
    // aiSolve(a,ai_b(i),c,ai_d(i),e,f,l,pi,u); 
      sum = 1/(e+f);

      b = ai_b(i);

//      d = ai_d(i);

      //      cout << "MATRIX " << i << endl;
      //      aiPrint(a,b,c,d,e,f);

      a *= sum;
      b *= sum;
      c += a*e;
      d += b*f;

      pi = c/d;

      _u = a + pi*b;
      
      sum = (double) 1.0/(1.0 + pi + _u);
      // ### change back to l
      l = sum; // ### tres questionable
      pi *= sum;
      if (pi < tol) pi = tol;
      // ### remove!
      //      u = _u*sum;

//      cout << "i) l = " << l << " pi = " << pi << " u = " << _u*sum << endl;

//////////////////////////////////////////////////////////////////////
      solution[i] = pi;
   
      l += solution(i);
    }

    u -= solution(tangible - 2);
//////////////////////////////////////////////////////////////////////

 // aiCalcLastParameters(tangible - 2, a,ai_b,c,ai_d,e,f, solution, l, u);
// ###
    //    if (l < tol) l = tol;

    i = tangible - 2;
    d = ai_d(i); // ###
    _col = &QC->getCol(i);

    cc = 0;
    for (rk=0; rk<_col->getEntries(); rk++) {
      k = _col->getRow(rk);
      if (k < i) cc += solution(k)*QC->getColElement(_col, rk);
    }
    c = cc/l;
    
    aa = aa + solution(i-1)*ai_b(i-1) - cc;
    a = aa/l;

    e = QCnn_1;
    f = 1 - (QCnn + e);

//////////////////////////////////////////////////////////////////////

//  aiSolve(a,ai_b(tangible - 2),c,ai_d(tangible - 2),e,f,l,pi,u); 

    b = ai_b(i);
//    d = ai_d(i);

//    aiPrint(a,b,c,d,e,f);

    sum = e+f;
    a /= sum;
    b /= sum;
    c += a*e;
    d += b*f;

    pi = c/d;
    u = a + pi*b;

    sum = (double) 1.0/(1.0 + pi + u);
    l = sum;
    pi *= sum;
    u *= sum;

    if (pi < tol) pi = tol;
    if (u < tol) u = tol;

//    cout << "n) l = " << l << " pi = " << pi << " u = " << u << endl;

//////////////////////////////////////////////////////////////////////

    solution[tangible - 2] = pi;
    solution[tangible - 1] = u;

    //cout << "solution(" << tangible - 2 << ") = " << pi << endl;
    //cout << "solution(" << tangible - 1 << ") = " << u << endl;
//    cout << "solution's one norm is " << solution.oneNorm() << endl;
    solution.normalise();
#ifdef DEBUG
    cout << "solution: " << solution << endl;
#endif
    iterations++;
#ifdef EXPERIMENT
    for (i=0; i<tangible; i++)
      normSol[i] = solution[i];
    normSol.normalise();
    Q->transMultiply(normSol, normRes);
    for (i=0; i<tangible; i++)
      normRes[i] -= normSol[i];
    cout << "iteration " << iterations << " infNorm=" << normRes.infNorm() << " twoNorm=" << normRes.twoNorm() << endl;
    cout << "EXPERIMENT " << iterations << " " << log10(normRes.twoNorm()) << endl;
#endif
    if (!(iterations % 5)) {
      maxDiff = 0;
      solNorm = 0;
      for (i = 0; i < tangible; i++) {
        if (fabs(solution(i)) > solNorm) solNorm = fabs(solution(i));
        if (fabs(solution(i)-last(i)) > maxDiff)
          maxDiff = fabs(solution(i) - last(i));
        last(i) = solution(i);
      }
      if (maxDiff != 0) convergence = maxDiff/solNorm;
        else convergence = 0;
    if (solRepStyle != NONE) {
      cout << setw(30) << iterations << "    | " << setw(34) << convergence << endl;
    }

    // ### PUT THESE DEBUGS BACK!
    // #ifdef DEBUG
      residInfNorm = 0;
      QC->transMultiply(solution, result);
      for (i = 0; i<tangible; i++) {
	result(i) -= solution(i);
	if (fabs(result(i)) > residInfNorm)
	  residInfNorm = fabs(result(i));
      }
      //cout << "average residual = " << result.oneNorm()/tangible << endl;
      //cout << "residual[0] = " << result[0] << endl;
      //cout << "residual[" << tangible - 1 << "] = " << result[tangible-1] << endl;

      //cout << "iteration " << iterations << " residual inf norm = " << residInfNorm << endl;
      // #endif
    }
  };

  if (convergence > accuracy)
    terminate("AI failed to converge within %d iterations", maxIterations);

  residInfNorm = 0;
  QC->transMultiply(solution, result);
  for (i = 0; i<tangible; i++) {
    result[i] -= solution(i);
    if (fabs(result(i)) > residInfNorm)
      residInfNorm = fabs(result(i));
  }

  if (solRepStyle != NONE)
    cout << "----------------------------------+---------------------------------------" << endl;

  cout.unsetf(ios::showpoint);
  checkViolation(checkVector(solution, residInfNorm));
  if (refine) {
    cout << "SteadyStateAnalyser (" << this << "): AI refine complete." << endl;
    return;
  }
  unScale(solution);
  solution.normalise();
  double timeUsed = timer->read();
  solution.output(steadyVectorFile); 
  #ifndef RELEASE
  solution.textOutput("ai.sol");
  #endif
  #ifndef SILENT 
  cout << "SteadyStateAnalyser (" << this << "): AI complete (" << timeUsed << " CPU seconds)." << endl;
  #endif
}





