////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

using namespace std;

#include <iostream>
#include <cassert>
//#define real double
#include "dyna.hpp"
#include "store.hpp"
#include "avl.hpp"

Store *AVLNode::store = NULL;

AVLNode::AVLNode(double *_index)
{
  #ifdef DEBUG
  cout << "AVLNode::AVLNode (" << *_index <<  ")" << endl;
  #endif
  index = _index;
  balance = 0;
  llink = rlink = NULL;
}

AVLNode::~AVLNode()
{
}

int AVLTree::height()
{
  return height(root);
}

int AVLTree::height(AVLNode *p)
{
  int height_left, height_right;
  if (!p) return 0;
  height_left = height(p->getLlink()) + 1;
  height_right = height(p->getRlink()) + 1;
  if (height_left>height_right) return height_left; else return height_right;
}

AVLNode *AVLTree::leftRotation(AVLNode *p)
{
  AVLNode *hold,*q;
  #ifdef DEBUG
  cout << "left rotation\n";
  #endif
  q = p->getRlink();
  assert(q);
  hold = q->getLlink();
  q->setLlink(p);
  p->setRlink(hold);
  return q;
}

AVLNode *AVLTree::rightRotation(AVLNode *p)
{
  AVLNode *hold,*q;

  #ifdef DEBUG
  cout << "right rotation\n";
  #endif
  q = p->getLlink();
  assert(q);
  hold = q->getRlink();
  q->setRlink(p);
  p->setLlink(hold);
  return q;
}

AVLNode *AVLTree::insert(double *index, double entry)
{
  AVLNode *fp,*p,*fya,*ya,*q,*s;
  int imbal;

// double entry = store->getEntry(index);
#ifdef DEBUG
  cout << "inserting index = " << index << " entry = " << entry << endl;
#endif

  if (!root) {
    root = new AVLNode(index);
    return root;
  }

  fp = fya = NULL;
  p = ya = root;

  /* ya now points to the youngest ancestor which may become unbalanced */
  /* fya points to the father of ya and fp points to the father of p */

  /* locate position in tree for new node */
  while (p) {
    if (p->getEntry() == entry) {
      return p;
    }

    if (entry < p->getEntry() ) q = p->getLlink(); else q = p->getRlink();
    if (q && q->getBalance() != 0) {
      fya = p;
      ya = q;
    }
    fp = p;
    p = q;
  }

  /* create new node */
  q = new AVLNode(index);

  /* insert new node into tree */
  if (entry < fp->getEntry()) fp->setLlink(q); else fp->setRlink(q);
  if (entry < ya->getEntry()) p = ya->getLlink(); else p = ya->getRlink();
  s = p;

  /* change balance of all nodes between youngest unbalanced ancestor
     and new node */
  while (p != q) {
    if (entry < p->getEntry()) {
      p->setBalance(1);
      p = p->getLlink();
    } else {
      p->setBalance(-1);
      p = p->getRlink();
    }
  }
  
  /* if tree unbalanced, q is the newly inserted node, ya is its youngest
     unbalanced ancestor, fya is the father of ya and s is the son of ya
     in the direction of the imbalance */

  #ifdef DEBUG
  cout << "ya->getEntry is " << ya->getEntry() << endl;
  #endif
  if (entry < ya->getEntry()) imbal = 1; else imbal = -1;
  if (ya->getBalance() == 0) {
    ya->setBalance(imbal);
    return q;
  }
  if (ya->getBalance() != imbal) {
    ya->setBalance(0);
    return q;
  }

  #ifdef DEBUG
  cout << "inserting " << entry << " imbal is " << imbal << endl;
  #endif
  if (s->getBalance() == imbal) {
    /* ya and s are unbalanced in the same direction */
    p = s;
    if (imbal == 1) rightRotation(ya);
      else leftRotation(ya);
    ya->setBalance(0);
    s->setBalance(0);
  } else {
    /* ya and s are unbalanced in opposite directions */
    if (imbal==1) {
      p = s->getRlink();
      leftRotation(s);
      ya->setLlink(p);
      rightRotation(ya);
    }
    else {
      p = s->getLlink();
      ya->setRlink(p);
      rightRotation(s);
      leftRotation(ya);
    }
    if (p->getBalance() == 0) {
      ya->setBalance(0);
      s->setBalance(0);
    }
    else
      if (p->getBalance() == imbal) {
        ya->setBalance(-imbal);
        s->setBalance(0);
      }
      else {
        ya->setBalance(0);
        s->setBalance(imbal);
      }
    p->setBalance(0);
  }

  if (!fya)
    root = p;
  else
    (ya == fya->getRlink() ) ? (fya->setRlink(p)) : (fya->setLlink(p));
  return q;
}

double *AVLTree::search(double entry)
{
  AVLNode *p = root;
  while (p) {
    if (entry == p->getEntry() ) {
      #ifdef DEBUG
      cout << "entry is " << entry << " p->getEntry is " << p->getEntry() << "(" << *(p->getIndex()) << ")" << endl;
      #endif
      return p->getIndex();
    }
    if (entry < p->getEntry() ) p = p->getLlink();
      else p = p->getRlink();
  }
  #ifdef DEBUG
  cout << "*** AVLTree:: search for entry " << entry << " failed!" << endl;
  #endif
  return NULL;
}

void AVLTree::destroy(AVLNode *p)
{
  if (!p) return;
  destroy(p->getLlink());
  destroy(p->getRlink());
  delete p;
}



