class Node {
protected:
  long component;
  int offspring;
  FastDynaArray<long> children;
public:
#ifdef DEBUG
  long magic;
#endif
  Node(int kids, StateMatrixElement *child);
  ~Node() {
#ifdef DEBUG
  magic = 0;
#endif
  }
  inline long getComponent() const { return component; }
  inline long getChild(int t) const { return children(t); }
  inline void setComponent(long _component) { component = _component; }
  inline int numberEnabled() const { return children.getLength(); }
};

class LocalParameters {
public:
  long v,w;
  int t;
};

class FunctionalAnalyser {
protected:
  FastDynaArray<Node *> state;
  FastDynaArray<long> stack, members, mapping;
  FastDynaArray<LocalParameters> local;
  void *core;

  Information *info;
  Timer *timer;
  BitField removed, final, transient;

  FastDynaArray<long> low, dfsNumber;

  long states, trans, components, height, level, dfn, maxLevel,
maxHeight, repInterval, transients, recurrent;
  ReportStyle repStyle;
  int transitions;

  void terminate(char *fmt, ...);
  void destroy();

  void SCompInitialise();
  void SCompDFS(long v);
  void SCompAnalyse(char*);
  void SCompFinish();

  inline void pushLocalVars(long _v, long _w, int _t) {
    local.check(level);
    local[level].v = _v;
    local[level].w = _w;    
    local[level].t = _t;
    level++;
    if (level>maxLevel) maxLevel = level;
  }

  inline void popLocalVars(long &_v, long  &_w, int &_t) {
    level--;
    #ifdef SAFE
    assert(level>=0);
    #endif
    _v = local[level].v;
    _w = local[level].w;
    _t = local[level].t;
  }

  inline void push(long n) {
//    cout << "Push: " << n << "\n";
    stack.check(height);
    stack[height] = n;
    height++;
    if (height>maxHeight) maxHeight = height;
  }

  inline long pop() {
    height--;
//    cout << "Pop: " << height << "\n";
    #ifdef SAFE
    assert(height>=0);
    #endif
    return stack[height];
  }

  inline long top() { return stack[height - 1]; }

public:
  int readStates();
  void findStronglyConnectedComponents(char*);
  int remap();
  int remapMatrix();
  int remapStates();

  FunctionalAnalyser(char*, char*);
  ~FunctionalAnalyser();
};

