////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

//#define BUFFER_SIZE 16384

using namespace std;

#include <iostream>
#include <cassert>
#include <cstring>
#include <cstdio>
#include "common.hpp"
#include "dyna.hpp"
#include "info.hpp"

Information::Information() {
  parse = generation = functional = steady = 0;
  deadlock = reducible = 0;
  tangible = trans = transitions = stateSize = 0;
  stateMeasures = countMeasures = 0;
  maxStates = maxIterations = maxChildren = genRepInterval = solRepInterval = 0;
  maxCPUTime = accuracy = relaxParameter = maxRate = 0.0;
  dynamic = 1;

  genRepStyle = solRepStyle = NONE;
  method = AUTOMATIC;

  transType = NULL; 
  countEstimator = stateEstimator = NULL;
  safeHash = 0;

  t_start = t_stop = t_step = 0.0;
  sources = targets = excludeds = 0;
  doSteady = 0;
  doPassage = 0;
  doTransient = 0;

  strcpy(checksum, "");
}

void Information::setStateName(int index, char *name)
{
  statem.check(index);
  statem[index] = new char[strlen(name) + 1];
  strcpy(statem[index], name);
}

void Information::setCountName(int index, char *name)
{
  countm.check(index);
  countm[index] = new char[strlen(name) + 1];
  strcpy(countm[index], name);
}

int Information::openFile(char *filename, FILE *&fp, char *access, char *buff) const
{
  fp = fopen(filename, access);
  if (!fp) {
    cout << "info.cxx Information (" << this << "): cannot open file '" << filename << "'.\n";
    return 0;
  }
  setvbuf(fp, buff, _IOFBF, BUFFER_SIZE);
  return 1;
}

int Information::closeFile(FILE *fp) const
{
  return fclose(fp);
}

int Information::output(char *filename) const
{
  FILE *nf;
  int n, length;
  #ifdef LOGGING
     cout << "info.cxx Information (" << this << "): writing data to '" << filename << "'.\n";
  #endif
  char buffer[BUFFER_SIZE];
  if (!openFile(filename, nf, "wb", buffer)) return 0; // ### ERROR HANDLER
  fwrite(&transitions, sizeof(int), 1, nf);
  fwrite(transType, sizeof(TransType), transitions, nf);
  fwrite(priority, sizeof(int), transitions, nf);
  fwrite(&stateMeasures, sizeof(int), 1, nf);
  fwrite(&countMeasures, sizeof(int), 1, nf);
  fwrite(&maxStates, sizeof(long), 1, nf);
  fwrite(&maxIterations, sizeof(long), 1, nf);
  fwrite(&genRepInterval, sizeof(long), 1, nf);
  fwrite(&solRepInterval, sizeof(long), 1, nf);
  fwrite(&maxCPUTime, sizeof(double), 1, nf);
  fwrite(&safeHash, sizeof(int), 1, nf);
  fwrite(&accuracy, sizeof(double), 1, nf);
  fwrite(&relaxParameter, sizeof(double), 1, nf);
  fwrite(&dynamic, sizeof(int), 1, nf);
  fwrite(&genRepStyle, sizeof(ReportStyle), 1, nf);
  fwrite(&solRepStyle, sizeof(ReportStyle), 1, nf);
  fwrite(&method, sizeof(SolutionMethod), 1, nf);
  fwrite(stateListFile, sizeof(char), MAX_LENGTH, nf);
  fwrite(steadyStateFile, sizeof(char), MAX_LENGTH, nf);
  fwrite(startVectorFile, sizeof(char), MAX_LENGTH, nf);
  fwrite(transitionMatrixFile, sizeof(char), MAX_LENGTH, nf);
  fwrite(stateEstimator, sizeof(int), stateMeasures, nf);
  fwrite(countEstimator, sizeof(int), countMeasures, nf);
  for (n=0; n<stateMeasures; n++) {
    length = strlen(statem(n)) + 1; // ### + 1???
    fwrite(&length, sizeof(int), 1, nf);
    fwrite(statem(n), sizeof(char), length, nf);
  }
  for (n=0; n<countMeasures; n++) {
    length = strlen(countm(n)) + 1; // ### + 1???
    fwrite(&length, sizeof(int), 1, nf);
    fwrite(countm(n), sizeof(char), length, nf);
  }
  length = strlen(checksum) + 1;
  fwrite(&length, sizeof(int), 1, nf);
  fwrite(checksum, sizeof(char), length, nf);

  fwrite(&t_start, sizeof(double), 1, nf);
  fwrite(&t_stop, sizeof(double), 1, nf);
  fwrite(&t_step, sizeof(double), 1, nf);

  fwrite(&sources, sizeof(long), 1, nf);
  fwrite(&targets, sizeof(long), 1, nf);
  fwrite(&excludeds, sizeof(long), 1, nf);

  fwrite(&doSteady, sizeof(int), 1, nf);
  fwrite(&doPassage, sizeof(int), 1, nf);
  fwrite(&doTransient, sizeof(int), 1, nf);

  return !closeFile(nf);
}

int Information::outputStateInfo(char *filename) const
{
  FILE *nf;
  #ifdef LOGGING
     cout << "info.cxx Information (" << this << "): writing state data to '" << filename << "'.\n";
  #endif
  char buffer[BUFFER_SIZE];
  if (!openFile(filename, nf, "wb", buffer)) return 0; // ### ERROR HANDLER
  fwrite(&tangible, sizeof(long), 1, nf);
  fwrite(&stateSize, sizeof(long), 1, nf);
  fwrite(&trans, sizeof(long), 1, nf);
  fwrite(&maxChildren, sizeof(int), 1, nf);
  fwrite(&maxRate, sizeof(double), 1, nf);
  fwrite(&parse, sizeof(int), 1, nf);
  fwrite(&generation, sizeof(int), 1, nf);
  fwrite(&functional, sizeof(int), 1, nf);
  fwrite(&steady, sizeof(int), 1, nf);
  fwrite(&reducible, sizeof(int), 1, nf);
  fwrite(&deadlock, sizeof(int), 1, nf);
  return !closeFile(nf);
}

int Information::input(char *filename) {
  FILE *nf;
  int n, length;

  #ifdef LOGGING
     cout << "info.cxx Information (" << this << "): reading data from '" << filename << "'.\n";
  #endif
  char buffer[BUFFER_SIZE];
  if (!openFile(filename, nf, "rb", buffer)) return 0; // ### ERROR HANDLER
  fread(&transitions, sizeof(int), 1, nf); // ERROR CHECK
//  cout << "-- transitions: " << transitions << "\n";
  transType = new TransType[transitions];
  priority = new int[transitions];
  fread(transType, sizeof(TransType), transitions, nf);
  fread(priority, sizeof(int), transitions, nf);
  fread(&stateMeasures, sizeof(int), 1, nf);
//  cout << "--- stateMeasures: " << stateMeasures << endl;
  fread(&countMeasures, sizeof(int), 1, nf);
//  cout << "--- countMeasures: " << countMeasures << endl;
  fread(&maxStates, sizeof(long), 1, nf);
//  cout << "--- maxStates: " << maxStates << endl;
  fread(&maxIterations, sizeof(long), 1, nf);
//  cout << "--- maxIterations: " << maxIterations << endl;
  fread(&genRepInterval, sizeof(long), 1, nf);
//  cout << "--- genRepInterval: " << genRepInterval << endl;
  fread(&solRepInterval, sizeof(long), 1, nf);
//  cout << "--- solRepInterval: " << solRepInterval << endl;
  fread(&maxCPUTime, sizeof(double), 1, nf);
  fread(&safeHash, sizeof(int), 1, nf);
//  cout << "--- maxCPUTime: " << maxCPUTime << endl;
  fread(&accuracy, sizeof(double), 1, nf);
//  cout << "--- accuracy: " << accuracy << endl;
  fread(&relaxParameter, sizeof(double), 1, nf);
//  cout << "--- relaxParameter: " << relaxParameter << endl;
  fread(&dynamic, sizeof(int), 1, nf);
//  cout << "--- dynamic: " << dynamic << endl;
  fread(&genRepStyle, sizeof(ReportStyle), 1, nf);
//  cout << "--- genRepStyle: " << int(genRepStyle) << endl;
  fread(&solRepStyle, sizeof(ReportStyle), 1, nf);
//  cout << "--- solRepStyle: " << int(solRepStyle) << endl;
  fread(&method, sizeof(SolutionMethod), 1, nf);
//  cout << "--- method: " << int(method) << endl;
  fread(stateListFile, sizeof(char), MAX_LENGTH, nf);
//  cout << "--- stateListFile: " << stateListFile << endl;
  fread(steadyStateFile, sizeof(char), MAX_LENGTH, nf);
//  cout << "--- steadyStateFile: " << steadyStateFile << endl;
  fread(startVectorFile, sizeof(char), MAX_LENGTH, nf);
//  cout << "--- startVectorfile: " << startVectorFile << endl;
  fread(transitionMatrixFile, sizeof(char), MAX_LENGTH, nf);
//  cout << "--- transitionMatrixFile: " << transitionMatrixFile << endl;
  stateEstimator = new int[stateMeasures];
  countEstimator = new int[countMeasures];
  fread(stateEstimator, sizeof(int), stateMeasures, nf);
  fread(countEstimator, sizeof(int), countMeasures, nf);
  for (n=0; n<stateMeasures; n++) {
    fread(&length, sizeof(int), 1, nf);
    statem.check(n);
    statem[n] = new char[length];
    fread(statem[n], sizeof(char), length, nf);
//    cout << "--- statem[" << n << "] = " << statem(n) << endl;
  }
  for (n=0; n<countMeasures; n++) {
    fread(&length, sizeof(int), 1, nf);
    countm.check(n);
    countm[n] = new char[length];
    fread(countm[n], sizeof(char), length, nf);
//    cout << "--- countm[" << n << "] = " << countm(n) << endl;
  }

  fread(&length, sizeof(int), 1, nf);
  fread(checksum, sizeof(char), length, nf);

#ifdef LOGGING
  cout << "checksum = " << checksum << endl;
#endif


  fread(&t_start, sizeof(double), 1, nf);
  fread(&t_stop, sizeof(double), 1, nf);
  fread(&t_step, sizeof(double), 1, nf);

  fread(&sources, sizeof(long), 1, nf);
  fread(&targets, sizeof(long), 1, nf);
  fread(&excludeds, sizeof(long), 1, nf);

  fread(&doSteady, sizeof(int), 1, nf);
  fread(&doPassage, sizeof(int), 1, nf);
  fread(&doTransient, sizeof(int), 1, nf);

  return !closeFile(nf);
}

/*
  fwrite(&tangible, sizeof(long), 1, nf);
  fwrite(&trans, sizeof(long), 1, nf);
  fwrite(&maxChildren, sizeof(int), 1, nf);
  fwrite(&parse, sizeof(int), 1, nf);
  fwrite(&generation, sizeof(int), 1, nf);
  fwrite(&functional, sizeof(int), 1, nf);
  fwrite(&steady, sizeof(int), 1, nf);
  fwrite(&reducible, sizeof(int), 1, nf);
*/

int Information::inputStateInfo(char *filename) {
  FILE *nf;

  #ifdef LOGGING
     cout << "info.cxx Information (" << this << "): reading state data from '" << filename << "'.\n";
  #endif
  char buffer[BUFFER_SIZE];
  if (!openFile(filename, nf, "rb", buffer)) return 0; // ### ERROR HANDLER
  fread(&tangible, sizeof(long), 1, nf);
  #ifdef LOGGING
     cout << "-- totalTangible: " << tangible << "\n";
  #endif
  fread(&stateSize, sizeof(long), 1, nf);
  fread(&trans, sizeof(long), 1, nf);
  #ifdef LOGGING
     cout << "-- totalTrans: " << trans << "\n";
  #endif
  fread(&maxChildren, sizeof(int), 1, nf);
  #ifdef LOGGING
     cout << "-- maxChildren: " << maxChildren << "\n";
  #endif
  fread(&maxRate, sizeof(double), 1, nf);
  fread(&parse, sizeof(int), 1, nf);
  #ifdef LOGGING
     cout << "-- parse: " << parse << "\n";
  #endif
  fread(&generation, sizeof(int), 1, nf);
  #ifdef LOGGING
     cout << "-- generation: " << generation << "\n";
  #endif
  fread(&functional, sizeof(int), 1, nf);
  #ifdef LOGGING
     cout << "-- functional: " << functional << "\n";
  #endif
  fread(&steady, sizeof(int), 1, nf);
  #ifdef LOGGING
     cout << "-- steady: " << steady << "\n";
  #endif
  fread(&reducible, sizeof(int), 1, nf);
  #ifdef LOGGING
     cout << "-- reducible: " << reducible << "\n";
  #endif
  fread(&deadlock, sizeof(int), 1, nf);
  return !closeFile(nf);
}


