class StateMeasure {
private:
  double sum, sum2;
  int values;
  FastDynaArray<double> dbnValues, dbnFrequency;
  int findIndex(double value) {
//    cout << "FIND INDEX" << endl;
    for (int n=0; n < values; n++)
      if (dbnValues(n) == value) return n;
    dbnValues.check(values);
    dbnValues[values] = value;
    dbnFrequency.check(values);
    dbnFrequency[values] = 0;
    return values++;
  }
public:
  StateMeasure() {
    sum = 0;
    sum2 = 0;
    values = 0;
  }
  inline double getSum() { return sum; }
  inline double getVariance() { return sum2-sum*sum; }
  inline double getStdDev() { return sqrt(sum2 - sum*sum); }
  inline double getCoeffOfVariation() { return getStdDev()/sum; }
  inline void setSum(double _sum) { sum = _sum; }
  inline int getValues() { return values; }
  inline void add(double value, double prob) {
//    cout << "ADD value = " << value << ", prob = " << prob << " BEGIN" << endl;
    sum += value*prob;
    sum2 += value*value*prob;
    int index = findIndex(value);
//    cout << "INDEX = " << index << endl;
    dbnFrequency[index] += prob;
//    cout << "ADD END" << endl;
  }
  inline double getFrequency(int n) { return dbnFrequency(n); }
  inline double getValue(int n) { return dbnValues(n); }
  int showDistribution();
};

class ValueListEntry {
private:
  double dbnValue, dbnFrequency;
  ValueListEntry *next;
public:
  ValueListEntry(double _dbnValue, double _dbnFrequency) {
    dbnValue = _dbnValue;
    dbnFrequency = _dbnFrequency;
    next = NULL;
  }
  inline double getDbnValue() { return dbnValue; }
  inline double getDbnFrequency() { return dbnFrequency; }
  void setNext(ValueListEntry *_next) { next = _next; }
  ValueListEntry *getNext() const { return next; }
};

class ValueList {
  ValueListEntry *front;
public:
  ValueList() {
    front = NULL;
  }
  void insert(double _dbnValue, double _dbnFrequency);
  void output() {
    ValueListEntry *p;
    for (p = front; p; p=p->getNext()) {
      cout.unsetf(ios::scientific);
      cout << "   " << setw(16) << p->getDbnValue();
      cout.setf(ios::scientific);
      cout << "    " << p->getDbnFrequency() << endl;
    }
  }
  void destroy() {
    ValueListEntry *p = front, *q;
    while (p) {
      q = p->getNext();
      delete p;
      p = q;
    }
    front = NULL;
  }
  ~ValueList() { destroy(); }
};

class PerformanceAnalyser
{
  Information *info;
  Timer *timer;
  void *core;

  void terminate(char *fmt, ...);
  void destroy();

  Vector<double> final;
  long states, trans, elements, vanish, tangible, genRepInterval;
  ReportStyle genRepStyle;
public:
  PerformanceAnalyser(char*);
  ~PerformanceAnalyser();
  int readFinalSolution();
  void computeMeasures();
};

