
// "gamma" is the implementation of the function for generating gamma-
// distributed random varriables from the file RandGamma.h distributed
// as part of the CLHEP package (http://wwwinfo.cern.ch/asd/lhc++/clhep/).
// Authors were John Marraffino & Gabriele Cosmo


double exponential_dist(double __lambda) {
    return -log(drand48())/__lambda; 
}

double uniform_dist(double __a, double __b) {
    double __value = drand48();
    return __a + (__value*(__b-__a));
}

double erlang_dist(double __lambda, int __n) {
    double __sum = 0.0;
    for(int __l=0; __l<__n; __l++)
      __sum += exponential_dist(__lambda);
    return __sum;
}

double gamma_dist(double __lambda, double __a) {
 
  static double __aaaa = -1.0, __aaaaa = -1.0, __b, __c, __d, __e, __r, __s, __si, __ss, __q0,
       __q1 = 0.0416666664, __q2 =  0.0208333723, __q3 = 0.0079849875,
       __q4 = 0.0015746717, __q5 = -0.0003349403, __q6 = 0.0003340332,
       __q7 = 0.0006053049, __q8 = -0.0004701849, __q9 = 0.0001710320,
       __a1 = 0.333333333,  __a2 = -0.249999949,  __a3 = 0.199999867,
       __a4 =-0.166677482,  __a5 =  0.142873973,  __a6 =-0.124385581,
       __a7 = 0.110368310,  __a8 = -0.112750886,  __a9 = 0.104089866,
       __e1 = 1.000000000,  __e2 =  0.499999994,  __e3 = 0.166666848,
       __e4 = 0.041664508,  __e5 =  0.008345522,  __e6 = 0.001353826,
       __e7 = 0.000247453;

  double __gds,__p,__q,__t,__sign_u,__u,__v,__w,__x;
  double __v1,__v2,__v12;

  // our modification is to use drand48() rather than CLHEP's flat()
  // to generate a uniformly distributed number between 0 and 1


  // Check for invalid input values

  if( __a <= 0.0 ) return (-1.0);
  if( __lambda <= 0.0 ) return (-1.0);

  if (__a < 1.0)
    {          // CASE A: Acceptance rejection algorithm gs
    __b = 1.0 + 0.36788794412 * __a;       // Step 1
    for(;;)
      {
       __p = __b * drand48();
       if (__p <= 1.0)
	  {                            // Step 2. Case gds <= 1
	    __gds = exp(log(__p) / __a);
	    if (log(drand48()) <= -__gds) return(__gds/__lambda);
	  }
       else
	 {                            // Step 3. Case gds > 1
	   __gds = - log ((__b - __p) / __a);
	   if (log(drand48()) <= ((__a - 1.0) * log(__gds))) return(__gds/__lambda);
	  }
      }
   }
 else
   {          // CASE B: Acceptance complement algorithm gd
    if (__a != __aaaa)
       {                               // Step 1. Preparations
	__aaaa = __a;
	__ss = __a - 0.5;
	__s = sqrt(__ss);
	__d = 5.656854249 - 12.0 * __s;
       }
                                              // Step 2. Normal deviate
    do {
      __v1 = 2.0 * drand48() - 1.0;
      __v2 = 2.0 * drand48() - 1.0;
      __v12 = __v1*__v1 + __v2*__v2;
    } while ( __v12 > 1.0 );
    __t = __v1*sqrt(-2.0*log(__v12)/__v12);
    __x = __s + 0.5 * __t;
    __gds = __x * __x;
    if (__t >= 0.0) return(__gds/__lambda);         // Immediate acceptance

    __u = drand48();            // Step 3. Uniform random number
    if (__d * __u <= __t * __t * __t) return(__gds/__lambda); // Squeeze acceptance

    if (__a != __aaaaa)
       {                               // Step 4. Set-up for hat case
	__aaaaa = __a;
	__r = 1.0 / __a;
	__q0 = ((((((((__q9 * __r + __q8) * __r + __q7) * __r + __q6) * __r + __q5) * __r + __q4) * __r + __q3) * __r + __q2) * __r + __q1) * __r;
	if (__a > 3.686)
	   {
	    if (__a > 13.022)
	       {
		__b = 1.77;
		__si = 0.75;
		__c = 0.1515 / __s;
	       }
	    else
	       {
		__b = 1.654 + 0.0076 * __ss;
		__si = 1.68 / __s + 0.275;
		__c = 0.062 / __s + 0.024;
	       }
	   }
	else
	   {
	    __b = 0.463 + __s - 0.178 * __ss;
	    __si = 1.235;
	    __c = 0.195 / __s - 0.079 + 0.016 * __s;
	   }
       }
    if (__x > 0.0)                       // Step 5. Calculation of q
       {
	__v = __t / (__s + __s);               // Step 6.
	if (fabs(__v) > 0.25)
	   {
	    __q = __q0 - __s * __t + 0.25 * __t * __t + (__ss + __ss) * log(1.0 + __v);
	   }
	else
	   {
	    __q = __q0 + 0.5 * __t * __t * ((((((((__a9 * __v + __a8) * __v + __a7) * __v + __a6) * __v + __a5) * __v + __a4) * __v + __a3) * __v + __a2) * __v + __a1) * __v;
	   }                // Step 7. Quotient acceptance
	if (log(1.0 - __u) <= __q) return(__gds/__lambda);
       }

    for(;;)
       {                    // Step 8. Double exponential deviate t
	do
	{
	 __e = -log(drand48());
	 __u = drand48();
	 __u = __u + __u - 1.0;
	 __sign_u = (__u > 0)? 1.0 : -1.0;
	 __t = __b + (__e * __si) * __sign_u;
	}
	while (__t <= -0.71874483771719);   // Step 9. Rejection of t
	__v = __t / (__s + __s);                  // Step 10. New q(t)
	if (fabs(__v) > 0.25)
	   {
	    __q = __q0 - __s * __t + 0.25 * __t * __t + (__ss + __ss) * log(1.0 + __v);
	   }
	else
	   {
	    __q = __q0 + 0.5 * __t * __t * ((((((((__a9 * __v + __a8) * __v + __a7) * __v + __a6) * __v + __a5) * __v + __a4) * __v + __a3) * __v + __a2) * __v + __a1) * __v;
	   }
	if (__q <= 0.0) continue;           // Step 11.
	if (__q > 0.5)
	   {
	    __w = exp(__q) - 1.0;
	   }
	else
	   {
	    __w = ((((((__e7 * __q + __e6) * __q + __e5) * __q + __e4) * __q + __e3) * __q + __e2) * __q + __e1) * __q;
	   }                    // Step 12. Hat acceptance
	if ( __c * __u * __sign_u <= __w * exp(__e - 0.5 * __t * __t))
	   {
	    __x = __s + 0.5 * __t;
	    return(__x*__x/__lambda);
	   }
       }
   }  
  }

double deterministic_dist(double __d) { return __d; }
double immediate_dist() { return 0; }
