class ComplexSparseMatrixRow {
protected:
  long entries;
  FastDynaArray<ComplexSteadyStateRowElement> element;
public:
  ComplexSparseMatrixRow(long n) { entries = n; element.setLength(n); }
  ComplexSparseMatrixRow() { entries = 0; }
  ~ComplexSparseMatrixRow() {} 
  long addElement(long column, complex<double> *index);
  long safeAddElement(long column, complex<double> *index);
  inline long getColumn(long index) const { return element(index).getColumn(); } 
  inline complex<double> *getIndex(long index) const { return element(index).getIndex(); }
  inline long getEntries() const { return entries; }
  inline void lock() { element.setLength(entries); }
};


class ComplexSparseMatrixCol {
protected:
  long entries;
  FastDynaArray<ComplexSteadyStateColElement> element;
public:
  ComplexSparseMatrixCol(long n) { entries = n; element.setLength(n); }
  ComplexSparseMatrixCol() { entries = 0; }
  ~ComplexSparseMatrixCol() {} 
  void setLength(long n) { element.setLength(n); }
  inline void addElement(long row, complex<double> *index) {
    element.check(entries);
    element[entries].setRow(row);
    element[entries].setIndex(index);
    entries++;
  }
  inline long safeAddElement(long row, complex<double> *index) {
    for (long n=0; n<entries; n++) {
      if (element(n).getRow() == row) {
	element(n).setIndex(index);
	return 0;
      }
    }
    element.check(entries);
    element[entries].setRow(row);
    element[entries].setIndex(index);
    entries++;
    return 1;
  }

  inline long getRow(long index) const { return element(index).getRow(); }
  inline complex<double> *getIndex(long index) const { return element(index).getIndex(); }
  inline long getEntries() const { return entries; }
  inline void lock() { element.setLength(entries); }
};


class SparseMatrixRow {
protected:
  long entries;
  FastDynaArray<SteadyStateRowElement> element;
public:
  SparseMatrixRow(long n) { entries = n; element.setLength(n); }
  SparseMatrixRow() { entries = 0; }
  ~SparseMatrixRow() {} 
  long addElement(long column, double *index);
  long safeAddElement(long column, double *index);
  inline long getColumn(long index) const { return element(index).getColumn(); } 
  inline double *getIndex(long index) const { return element(index).getIndex(); }
  inline long getEntries() const { return entries; }
  inline void lock() { element.setLength(entries); }
  inline void setLength(long n) { element.setLength(n); }
  inline void remove() { element.destroy(); entries = 0;}
};

class SparseMatrixCol {
protected:
  long entries;
  FastDynaArray<SteadyStateColElement> element;
public:
  SparseMatrixCol(long n) { entries = n; element.setLength(n); }
  SparseMatrixCol() { entries = 0; }
  ~SparseMatrixCol() {} 
  void setLength(long n) { element.setLength(n); }
  inline void addElement(long row, double *index) {
    element.check(entries);
    element[entries].setRow(row);
    element[entries].setIndex(index);
    entries++;
  }
  inline long safeAddElement(long row, double *index) {
    for (long n=0; n<entries; n++) {
      if (element(n).getRow() == row) {
	element(n).setIndex(index);
	return 0;
      }
    }
    element.check(entries);
    element[entries].setRow(row);
    element[entries].setIndex(index);
    entries++;
    return 1;
  }

  inline long getRow(long index) const { return element(index).getRow(); }
  inline double *getIndex(long index) const { return element(index).getIndex(); }
  inline long getEntries() const { return entries; }
  inline void lock() { element.setLength(entries); }
  inline void remove() { element.destroy(); entries=0;}
};


class StorelessSparseMatrixRow {
protected:
  long entries;
  FastDynaArray<StorelessSteadyStateRowElement> element;
public:
  StorelessSparseMatrixRow(long n) { entries = n; element.setLength(n); }
  StorelessSparseMatrixRow() { entries = 0; }
  ~StorelessSparseMatrixRow() {} 

  inline void shuffleUp(long col) {
    for(int n=0; n<entries; n++) {
      if(element[n].getColumn()==col) {
	for (int k=n; k<entries-1; k++) {
	  element[k].setColumn(element[k+1].getColumn());
	  element[k].setValue(element[k+1].getValue());
	}
        entries--;
	element.setLength(entries);       
	return;
      }
    }
    cout << "warning StorelessSparseMatrixRow::shuffleUp(" << col << ") not found!!" << endl;
  }


  inline long addElement(long column, double value) {
    element.check(entries);
    element[entries].setColumn(column);
    element[entries].setValue(value);
    return entries++;
  }

  inline long safeAddElement(long column, double value) {
    for (long n=0; n<entries; n++)  
      if (element(n).getColumn() == column) {
	element(n).setValue(value);
	return 0;
      }
    element.check(entries);
    element[entries].setColumn(column);
    element[entries].setValue(value);
    entries++;
    return 1;
  }

  inline long getColumn(long index) const { return element(index).getColumn(); } 
  inline double getValue(long index) const { return element(index).getValue(); }
  inline long getEntries() const { return entries; }
  inline void lock() { element.setLength(entries); }
  inline void setLength(long n) { element.setLength(n); }
  inline void remove() { element.destroy(); entries = 0;}
};


class StorelessSparseMatrixCol {
protected:
  long entries;
  FastDynaArray<StorelessSteadyStateColElement> element;
public:
  StorelessSparseMatrixCol(long n) { entries = n; element.setLength(n); }
  StorelessSparseMatrixCol() { entries = 0; }
  ~StorelessSparseMatrixCol() {} 
  void setLength(long n) { element.setLength(n); }

  inline void shuffleUp(long row) {
    for(int n=0; n<entries; n++) {
      if(element[n].getRow()==row) {
	for (int k=n; k<entries-1; k++) {
	  element[k].setRow(element[k+1].getRow());
	  element[k].setValue(element[k+1].getValue());
	}
        entries--;
	element.setLength(entries);       
	return;
      }
    }
    cout << "warning StorelessSparseMatrixCol::shuffleUp(" << row << ") not found!!" << endl;
  }


  inline long addElement(long row, double value) {
    element.check(entries);
    element[entries].setRow(row);
    element[entries].setValue(value);
    return entries++;
  }
  inline long safeAddElement(long row, double value) {
    for (long n=0; n<entries; n++) {
      if (element(n).getRow() == row) {
	element(n).setValue(value);
	return 0;
      }
    }
    element.check(entries);
    element[entries].setRow(row);
    element[entries].setValue(value);
    entries++;
    return 1;
  }

  inline long getRow(long index) const { return element(index).getRow(); }
  inline double getValue(long index) const { return element(index).getValue(); }
  inline long getEntries() const { return entries; }
  inline void lock() { element.setLength(entries); }
  inline void remove() { element.destroy(); entries=0;}
};


class IntegerStorelessSparseMatrixRow {
protected:
  long entries;
  FastDynaArray<IntegerStorelessSteadyStateRowElement> element;
public:
  IntegerStorelessSparseMatrixRow(long n) { entries = n; element.setLength(n); }
  IntegerStorelessSparseMatrixRow() { entries = 0; }
  ~IntegerStorelessSparseMatrixRow() {} 



  inline long addElement(long column, int value) {
    element.check(entries);
    element[entries].setColumn(column);
    element[entries].setValue(value);
    return entries++;
  }

  inline long safeAddElement(long column, int value) {
    for (long n=0; n<entries; n++)  
      if (element(n).getColumn() == column) {
	element(n).setValue(value);
	return 0;
      }
    element.check(entries);
    element[entries].setColumn(column);
    element[entries].setValue(value);
    entries++;
    return 1;
  }

  inline long getColumn(long index) const { return element(index).getColumn(); } 
  inline int getValue(long index) const { return element(index).getValue(); }
  inline long getEntries() const { return entries; }
  inline void lock() { element.setLength(entries); }
  inline void setLength(long n) { element.setLength(n); }
  inline void remove() { element.destroy(); entries = 0;}
};


class IntegerStorelessSparseMatrixCol {
protected:
  long entries;
  FastDynaArray<IntegerStorelessSteadyStateColElement> element;
public:
  IntegerStorelessSparseMatrixCol(long n) { entries = n; element.setLength(n); }
  IntegerStorelessSparseMatrixCol() { entries = 0; }
  ~IntegerStorelessSparseMatrixCol() {} 
  void setLength(long n) { element.setLength(n); }
  inline long addElement(long row, int value) {
    element.check(entries);
    element[entries].setRow(row);
    element[entries].setValue(value);
    return entries++;
  }
  inline long safeAddElement(long row, int value) {
    for (long n=0; n<entries; n++) {
      if (element(n).getRow() == row) {
	element(n).setValue(value);
	return 0;
      }
    }
    element.check(entries);
    element[entries].setRow(row);
    element[entries].setValue(value);
    entries++;
    return 1;
  }

  inline long getRow(long index) const { return element(index).getRow(); }
  inline int getValue(long index) const { return element(index).getValue(); }
  inline long getEntries() const { return entries; }
  inline void lock() { element.setLength(entries); }
  inline void remove() { element.destroy(); entries=0;}
};



#define ID_PASSAGETIMEMATRIX "PassageTimeMatrix (" << this << "): "

class PassageTimeMatrix {
protected:
  long rows, cols, elements;
  FastDynaArray<SparseMatrixRow> row;
  AVLTree *tree;
  Store *store;
  //  FastDynaArray< _complex_ > diag, rhs;
public:
  PassageTimeMatrix(long r, long c) {
    cout << ID_PASSAGETIMEMATRIX << "constructing " << r << " rows and " << c << " columns" << endl;
    rows = r;
    cols = c;
    row.setLength(rows);
    elements = 0;
    store = new Store;
    tree = new AVLTree(store);
    cout << ID_PASSAGETIMEMATRIX << "constructed with " << r << " rows and " << c << " columns" << endl;
    
  }
  ~PassageTimeMatrix() {
    cout << ID_PASSAGETIMEMATRIX << "destructing (" << store->getElements() << " doubles used)...";
    delete store;
    delete tree;
  }
  
  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getElements() const { return elements; }
  inline SparseMatrixRow &getRow(long index) { return row(index); }

  inline double *insert(double _entry) {
#ifdef VERBOSE
    cout << ID_PASSAGETIMEMATRIX << "inserting " << _entry << "..." << endl;
#endif
    double *_index = tree->search(_entry);
    if (_index == NULL) {
      _index = store->insert(_entry);
      tree->insert(_index, *_index );
      elements++;
    }
#ifdef VERBOSE
    cout << ID_PASSAGETIMEMATRIX << "confirming " << _index << "(" << *_index << ")" << endl;
#endif

    return _index;
  }  
  
  _complex_ dot( FastDynaArray< _complex_ > &one, FastDynaArray< _complex_ > &two );

  void SORIteration(FastDynaArray < _complex_ > &x, FastDynaArray < _complex_ > &diagonal, FastDynaArray<_real_> &rhs, _real_ omega);

  void SORIteration(FastDynaArray < _real_ > &x, FastDynaArray < _real_ > &diagonal, FastDynaArray<_real_> &rhs, _real_ omega);

  void calcResidual(FastDynaArray < _complex_ > &x, FastDynaArray < _complex_ > &diagonal, FastDynaArray<_real_> &rhs, FastDynaArray< _complex_ > &r);

  void multiply( FastDynaArray< _complex_ > &x, FastDynaArray< _complex_ > &diagonal, FastDynaArray < _complex_ > &result);

  inline void addElement(long _row, long _col, double _entry) {
#ifdef VERBOSE
    cout << ID_PASSAGETIMEMATRIX << "adding element (" << _row << ", " << _col << ", " << _entry << ")" << endl;
#endif
    double *_index = insert(_entry);    
#ifdef VERBOSE
    cout << ID_PASSAGETIMEMATRIX << "adding element to row " << _row << " (total rows = " << row.getLength() << ")" << endl;
#endif
    row[_row].addElement(_col, _index);
#ifdef VERBOSE
    cout << ID_PASSAGETIMEMATRIX << "element (" << _col << ", " << *_index << ") added" << endl;
#endif
  }

  inline double getElement(long _row, long _col) const {
    for (long n=0; n<row(_row).getEntries(); n++)
      if (row(_row).getColumn(n) == _col) return *(row(_row).getIndex(n));
    return 0.0;     
  }

  inline double getRowElement(SparseMatrixRow *_row, long _index) const {
    return *(_row->getIndex(_index));
  }
};

#define ID_CPTMX "ComplexPassageTimeMatrix (" << this << "): "

class ComplexPassageTimeMatrix {
protected:
  long rows, cols, elements;
  FastDynaArray<ComplexSparseMatrixRow> row;
  AVLTreeComplex *tree;
  ComplexStore *store;
  //  FastDynaArray< _complex_ > diag, rhs;
public:
  ComplexPassageTimeMatrix(long r, long c) {
    cout << ID_CPTMX << "constructing " << r << " rows and " << c << " columns" << endl;
    rows = r;
    cols = c;
    row.setLength(rows);
    elements = 0;
    store = new ComplexStore;
    tree = new AVLTreeComplex(store);
    cout << ID_CPTMX << "constructed with " << r << " rows and " << c << " columns" << endl;
    
  }
  ~ComplexPassageTimeMatrix() {
    cout << ID_CPTMX << "destructing (" << store->getElements() << " doubles used)...";
    delete store;
    delete tree;
  }
  
  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getElements() const { return elements; }
  inline ComplexSparseMatrixRow &getRow(long index) { return row(index); }

  inline complex<double> *insert(complex<double> _entry) {
#ifdef VERBOSE
    cout << ID_CPTMX << "inserting " << _entry << "..." << endl;
#endif
    complex<double> *_index = tree->search(_entry);
    if (_index == NULL) {
      _index = store->insert(_entry);
      tree->insert(_index, *_index );
      elements++;
    }
#ifdef VERBOSE
    cout << ID_CPTMX << "confirming " << _index << "(" << *_index << ")" << endl;
#endif

    return _index;
  }  
  
  _complex_ dot( FastDynaArray< _complex_ > &one, FastDynaArray< _complex_ > &two );

  void SORIteration(FastDynaArray < _complex_ > &x, FastDynaArray<_complex_> &diagonal, FastDynaArray<_complex_> &rhs, _real_ omega);

  //  void SORIteration(FastDynaArray < _real_ > &x, FastDynaArray < _real_ > &diagonal, FastDynaArray<_real_> &rhs, _real_ omega);

  void calcResidual(FastDynaArray < _complex_ > &x, FastDynaArray<_complex_> &rhs, FastDynaArray< _complex_ > &r);

  void multiply( FastDynaArray< _complex_ > &x, FastDynaArray < _complex_ > &result);

  inline void addElement(long _row, long _col, complex<double> _entry) {
#ifdef VERBOSE
    cout << ID_CPTMX << "adding element (" << _row << ", " << _col << ", " << _entry << ")" << endl;
#endif
    complex<double> *_index = insert(_entry);    
#ifdef VERBOSE
    cout << ID_CPTMX << "adding element to row " << _row << " (total rows = " << row.getLength() << ")" << endl;
#endif
    row[_row].addElement(_col, _index);
#ifdef VERBOSE
    cout << ID_CPTMX << "element (" << _col << ", " << *_index << ") added" << endl;
#endif
  }
  inline void safeAddElement(long _row, long _col, complex<double> _entry) {
#ifdef VERBOSE
    cout << ID_CPTMX << "adding element (" << _row << ", " << _col << ", " << _entry << ")" << endl;
#endif
    complex<double> *_index = insert(_entry);    
#ifdef VERBOSE
    cout << ID_CPTMX << "adding element to row " << _row << " (total rows = " << row.getLength() << ")" << endl;
#endif
    row[_row].safeAddElement(_col, _index);
#ifdef VERBOSE
    cout << ID_CPTMX << "element (" << _col << ", " << *_index << ") added" << endl;
#endif
  }

  inline complex<double> getElement(long _row, long _col) const {
    for (long n=0; n<row(_row).getEntries(); n++)
      if (row(_row).getColumn(n) == _col) return *(row(_row).getIndex(n));
    return 0.0;     
  }

  inline complex<double> getRowElement(ComplexSparseMatrixRow *_row, long _index) const {
    return *(_row->getIndex(_index));
  }
};

class ComplexPassageTimeMatrixWithCol {
protected:
  long rows, cols, elements, calls;
  FastDynaArray<ComplexSparseMatrixCol> col;

  AVLTreeComplex *tree;
  ComplexStore *store;
public:
  ComplexPassageTimeMatrixWithCol(long r, long c) {
#ifdef VERBOSE
    cout << "SparseMatrix (" << this << "): constructing " << r << " rows and " << c << " columns...\n";
  #endif
    rows = r;
    cols = c;
    calls = 0;
    col.setLength(cols);
    store = new ComplexStore;
    tree = new AVLTreeComplex(store);
#ifdef VERBOSE
    cout << "ComplexPassageTimeMatrixWithCol (" << this << "): constructed with " << r << " rows and " << c << " columns." << endl;
  #endif
  }
  ~ComplexPassageTimeMatrixWithCol();
  void statistics();
  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getCalls() const { return calls; }
  inline long getElements() const { return elements; }
  inline ComplexSparseMatrixCol &getCol(long index) { return col(index); }

  inline complex<double> *insert(complex<double> _entry) {
    #ifdef DEBUG
    cout << "insert: searching for " << _entry << endl;
    #endif
    calls++;
    complex<double> *_index = tree->search(_entry);
    if (_index == NULL) {
      _index = store->insert(_entry);
      tree->insert(_index, *_index );
/*      // ###
      assert(_index == tree->search(_entry));*/
      elements++;
    }
    return _index;
  }

  inline void addElement(long _row, long _column, complex<double> _entry) {
    complex<double> *_index = insert(_entry);    
    col[_column].addElement(_row, _index);
  }

  inline complex<double> getElement(long _row, long _column) const {
    for (long n=0; n<col(_column).getEntries(); n++)
      if (col(_column).getRow(n) == _row) return *(col(_column).getIndex(n));
    return 0.0;     
  }

  inline complex<double> getColElement(SparseMatrixCol *_col, long _index) const {
    return *(_col->getIndex(_index));
  }

  void transMultiply(FastDynaArray < _complex_ > &x, FastDynaArray < _complex_ > &result, int rank=0) const;
  void transMultiplyWithUnderflowDetection(FastDynaArray < _complex_ > &x, FastDynaArray < _complex_ > &result, int rank=0) const;
  //  void preMultiply(Vector<double> &x, Vector<double> &result) const;
  /*  void calcResidual(Vector<double> &x, Vector <double> &result) const;
  void SORIteration(Vector<double> &x, double omega) const;
  void safeSORIteration(Vector<double> &x, double omega) const;
  void twinSORIteration(Vector<double> &x1, double omega1, Vector<double>
&x2, double omega2) const;
  void calcResidualNorm(Vector<double> &x, double &xNorm, double &xInfNorm, Vector<double> &last, double &xDiffNorm) const;
  void calcResidualNormAI(Vector<double> &x, double &xNorm, double &xInfNorm, Vector<double> &last, double &xDiffNorm) const;
  void twinCalcResidualNorm(Vector<double> &x1, double &x1Norm, double &x1InfNorm,
Vector<double> &x2, double &x2Norm, double &x2InfNorm, Vector<double> &last,
double &x1DiffNorm, double &x2DiffNorm) const;

void print();*/
};

class SparseMatrix {
protected:
  long rows, cols, elements, calls;
  FastDynaArray<SparseMatrixCol> col;
  AVLTree *tree;
  Store *store;
public:
  SparseMatrix(long r, long c);
  ~SparseMatrix();
  void statistics();
  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getCalls() const { return calls; }
  inline long getElements() const { return elements; }
  inline SparseMatrixCol &getCol(long index) { return col(index); }

  inline double *insert(double _entry) {
    #ifdef DEBUG
    cout << "insert: searching for " << _entry << endl;
    #endif
    calls++;
    double *_index = tree->search(_entry);
    if (_index == NULL) {
      _index = store->insert(_entry);
      tree->insert(_index, *_index );
/*      // ###
      assert(_index == tree->search(_entry));*/
      elements++;
      //      cout << "store now has " << elements  << " elements" << endl;
    }
    return _index;
  }

  inline void addElement(long _row, long _column, double _entry) {
    double *_index = insert(_entry);    
    col[_column].addElement(_row, _index);
  }

  inline double getElement(long _row, long _column) const {
    for (long n=0; n<col(_column).getEntries(); n++)
      if (col(_column).getRow(n) == _row) return *(col(_column).getIndex(n));
    return 0.0;     
  }

  inline double getColElement(SparseMatrixCol *_col, long _index) const {
    return *(_col->getIndex(_index));
  }

  int invert();

  void transMultiply(Vector<double> &x, Vector <double> &result) const;
  void preMultiply(Vector<double> &x, Vector<double> &result) const;
  void calcResidual(Vector<double> &x, Vector <double> &result) const;
  void SORIteration(Vector<double> &x, double omega) const;
  void safeSORIteration(Vector<double> &x, double omega) const;
  void twinSORIteration(Vector<double> &x1, double omega1, Vector<double>
&x2, double omega2) const;
  void calcResidualNorm(Vector<double> &x, double &xNorm, double &xInfNorm, Vector<double> &last, double &xDiffNorm) const;
  void calcResidualNormAI(Vector<double> &x, double &xNorm, double &xInfNorm, Vector<double> &last, double &xDiffNorm) const;
  void twinCalcResidualNorm(Vector<double> &x1, double &x1Norm, double &x1InfNorm,
Vector<double> &x2, double &x2Norm, double &x2InfNorm, Vector<double> &last,
double &x1DiffNorm, double &x2DiffNorm) const;

  void print();
};

class HyperSparseMatrix {
protected:
  long rows, cols, elements;
  FastDynaArray<HyperSparseElement> element;
  //AVLTree *tree;
  //  Store *store;
public:
  HyperSparseMatrix(long r, long c);
  ~HyperSparseMatrix();
  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getElements() const { return elements; }
  inline HyperSparseElement &getElement(long index) { return element(index); }

  /*  inline double *insert(double _entry) {
    #ifdef DEBUG
    cout << "insert: searching for " << _entry << endl;
    #endif
    calls++;
    double *_index = tree->search(_entry);
    if (_index == NULL) {
      _index = store->insert(_entry);
      tree->insert(_index, *_index );
      elements++;
    }
    return _index;
    }*/

  inline void addElement(long _row, long _column, double _entry) {
    //    double *_index = insert(_entry);    
    //    row[_row].addElement(_column, _index);
    element.check(elements);
    element[elements].setRow(_row);
    element[elements].setColumn(_column);
    element[elements].setWeight(_entry);
    elements++;
  }

  inline double getElement(long _row, long _column) const {
    for (long n=0; n<elements; n++)
      if (element(n).getRow() == _row && element(n).getColumn() == _column) 
	return element(n).getWeight();
    return 0.0;     
  }

  void multiply(Vector<double> &x, Vector <double> &result) const;
  void transMultiply(Vector<double> &x, Vector <double> &result) const;
};


class ComplexHyperSparseMatrix {
protected:
  long rows, cols, elements;
  FastDynaArray<ComplexHyperSparseElement> element;
  //AVLTree *tree;
  //  Store *store;
public:
  ComplexHyperSparseMatrix(long r, long c);
  ~ComplexHyperSparseMatrix();
  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getElements() const { return elements; }
  inline ComplexHyperSparseElement &getElement(long index) { return element(index); }

  /*  inline double *insert(double _entry) {
    #ifdef DEBUG
    cout << "insert: searching for " << _entry << endl;
    #endif
    calls++;
    double *_index = tree->search(_entry);
    if (_index == NULL) {
      _index = store->insert(_entry);
      tree->insert(_index, *_index );
      elements++;
    }
    return _index;
    }*/

  inline void addElement(long _row, long _column, complex<double> _entry) {
    //    double *_index = insert(_entry);    
    //    row[_row].addElement(_column, _index);
    element.check(elements);
    element[elements].setRow(_row);
    element[elements].setColumn(_column);
    element[elements].setWeight(_entry);
    elements++;
  }

  inline complex<double> getElement(long _row, long _column) const {
    for (long n=0; n<elements; n++)
      if (element(n).getRow() == _row && element(n).getColumn() == _column) 
	return element(n).getWeight();
    return 0.0;     
  }

  void multiply(FastDynaArray<_complex_> &x, FastDynaArray<_complex_> &result) const;
  void transMultiply(FastDynaArray<_complex_> &x, FastDynaArray<_complex_> &result, int rank = 0) const;
};


class SparseMatrixWithRow {
protected:
  long rows, cols, elements, calls;
  FastDynaArray<SparseMatrixRow> row;
  AVLTree *tree;
  Store *store;
public:
  SparseMatrixWithRow(long r, long c);
  ~SparseMatrixWithRow();
  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getCalls() const { return calls; }
  inline long getElements() const { return elements; }
  inline SparseMatrixRow &getRow(long index) { return row(index); }

  inline double *insert(double _entry) {
    #ifdef DEBUG
    cout << "insert: searching for " << _entry << endl;
    #endif
    calls++;
    double *_index = tree->search(_entry);
    if (_index == NULL) {
      _index = store->insert(_entry);
      tree->insert(_index, *_index );
      elements++;
      //      cout << "store now has " << elements << " elements" << endl;
    }
    return _index;
  }

  inline void addElement(long _row, long _column, double _entry) {
    double *_index = insert(_entry);    
    row[_row].addElement(_column, _index);
  }

  inline int safeAddElement(long _row, long _column, double _entry) {
    double weight = getElement(_row, _column);
    double *_index = insert(_entry + weight);    
    return row[_row].safeAddElement(_column, _index);
  }

  inline double getElement(long _row, long _column) const {
    for (long n=0; n<row(_row).getEntries(); n++)
      if (row(_row).getColumn(n) == _column) return *(row(_row).getIndex(n));
    return 0.0;     
  }

  inline double getRowElement(SparseMatrixRow *_row, long _index) const {
    return *(_row->getIndex(_index));
  }

  void multiply(Vector<double> &x, Vector <double> &result) const;
  void inner(Vector<double> &x, Vector <double> &fixed) const;
  void innerSOR(Vector<double> &x, Vector <double> &fixed, double omega) const;
  void innerSOR(Vector<double> &x, Vector <double> &fixed, double &conv, double omega) const;
  void innerSOR(Vector<double> &x, Vector <double> &fixed, int lower, int upper, double omega) const;
  void innerSOR(Vector<double> &x, Vector <double> &fixed, double &maxSol, double &maxDiff, int lower, int upper, double omega) const;
  void integrity(char *message) const;
  void inner(Vector<double> &x, Vector <double> &fixed, double &conv) const;
  void print();
};


class SparseMatrixWithRowColumn : public SparseMatrix {
protected:
  FastDynaArray<SparseMatrixRow> row;
  int non_zeros;

public:
  SparseMatrixWithRowColumn(long r, long c) : SparseMatrix (r,c) {
//    cout << "SMWRC!" << endl;
    row.setLength(rows);
//    cout << "SMWRC constructed!" << endl;
    non_zeros = 0;
  }

  void removeState(int target) {

    cout << "target for removal=" << target << endl;
    //    row[target].setLength(0);
    //    col[target].setLength(0);

    row[target].remove();
    col[target].remove();

  }

  void multiply(Vector<double> &x, Vector<double> &result) const;
  inline void addElement(long _row, long _column, double _entry) {
    double *_index = insert(_entry);    
    row[_row].addElement(_column, _index);
    col[_column].addElement(_row, _index);    
    non_zeros++;
  }

  inline int safeAddElement(long _row, long _column, double _entry) {
    double weight = getElement(_row, _column);
    if (weight != 0)
      non_zeros++;
    double *_index = insert(_entry + weight);    
    int result = row[_row].safeAddElement(_column, _index);
    col[_column].safeAddElement(_row, _index);
    return result;
  }

  inline SparseMatrixRow &getRow(long index) { return row(index); }

  inline double getRowElement(long _row, long _index) const {
    return *(row(_row).getIndex(_index));
  }

  inline double getRowElement(SparseMatrixRow *_row, long _index) const {
//    double *index = _row->getIndex(_index);
    return *(_row->getIndex(_index));
  }

/*  inline void setElement(long _row, long _column, double _entry) {
    long _index = insert(_entry);
    row[_row].addElement(_column, _index);
    col[_column].addElement(_row, _index);
  }*/

  void print();
};

class DirectMatrix : public SparseMatrix { 
protected:
  FastDynaArray<SparseMatrixRow> row;
public:
  DirectMatrix(long r, long c) : SparseMatrix (r,c) {
    row.setLength(rows);
  }
  inline SparseMatrixRow &getRow(long index) { return row(index); }
  inline double *insert(double _entry) {
    calls++;
    double *_index = store->insert(_entry);
    elements++;
    return _index;
  }

  inline void set(double *_index, double _entry) {
    *_index = _entry;
//    store->setEntry(_index, _entry);
  }

  inline void addElement(long _row, long _column, double _entry) {
    double *_index = insert(_entry);
    row[_row].addElement(_column, _index);
    col[_column].addElement(_row, _index);
  }

  inline double getElement(long _row, long _column) const {
    for (long n=0; n<row(_row).getEntries(); n++)
      if (row(_row).getColumn(n) == _column) return *(row(_row).getIndex(n));
    return 0.0;     
  }

  inline double *getIndex(long _row, long _column) const
  {
    for (long n=0; n<row(_row).getEntries(); n++)
      if (row(_row).getColumn(n) == _column) return row(_row).getIndex(n);
    return NULL;
  }  

  inline void setElement(long _row, long _column, double _entry) {
    double *_index = getIndex(_row, _column);
    if (_index == NULL) {
      if (!_entry) return;
      _index = insert(_entry);
      row[_row].addElement(_column, _index);
      col[_column].addElement(_row, _index);
    } else
      set(_index, _entry);
  }

  inline void addToElement(long _row, long _column, double _entry) {
    double *_index = getIndex(_row, _column);
    if (_index == NULL) {
      if (!_entry) return;
      _index = insert(_entry);
      row[_row].addElement(_column, _index);
      col[_column].addElement(_row, _index);
    } else
      set(_index, _entry + (*_index) );
  }

  inline void divideColElement(SparseMatrixCol *_col, long index, double factor)
  {
    double *_index = _col->getIndex(index);
    set(_index, (*_index)/factor);
  }

  inline double getColElement(SparseMatrixCol *_col, long _index) const {
    double *index = _col->getIndex(_index);
    return (*index);
  }

  inline double getRowElement(SparseMatrixRow *_row, long _index) const {
    double *index = _row->getIndex(_index);
    return (*index);
  }

  void print();
};


/*
class SparseMatrix {
protected:
  long rows, cols, elements, calls;
  FastDynaArray<SparseMatrixCol> col;
  AVLTree *tree;
  Store *store;
public:
  SparseMatrix(long r, long c);
  ~SparseMatrix();
  void statistics();
  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getCalls() const { return calls; }
  inline long getElements() const { return elements; }
  inline SparseMatrixCol &getCol(long index) { return col(index); }

  inline double *insert(double _entry) {
    #ifdef DEBUG
    cout << "insert: searching for " << _entry << endl;
    #endif
    calls++;
    double *_index = tree->search(_entry);
    if (_index == NULL) {
      _index = store->insert(_entry);
      tree->insert(_index, *_index );
      elements++;
      cout << "store now has " << elements  << " elements" << endl;
    }
    return _index;
  }

  inline void addElement(long _row, long _column, double _entry) {
    double *_index = insert(_entry);    
    col[_column].addElement(_row, _index);
  }

  inline double getElement(long _row, long _column) const {
    for (long n=0; n<col(_column).getEntries(); n++)
      if (col(_column).getRow(n) == _row) return *(col(_column).getIndex(n));
    return 0.0;     
  }

  inline double getColElement(SparseMatrixCol *_col, long _index) const {
    return *(_col->getIndex(_index));
  }

  int invert();

  void transMultiply(Vector<double> &x, Vector <double> &result) const;
  void preMultiply(Vector<double> &x, Vector<double> &result) const;
  void calcResidual(Vector<double> &x, Vector <double> &result) const;
  void SORIteration(Vector<double> &x, double omega) const;
  void safeSORIteration(Vector<double> &x, double omega) const;
  void twinSORIteration(Vector<double> &x1, double omega1, Vector<double>
&x2, double omega2) const;
  void calcResidualNorm(Vector<double> &x, double &xNorm, double &xInfNorm, Vector<double> &last, double &xDiffNorm) const;
  void calcResidualNormAI(Vector<double> &x, double &xNorm, double &xInfNorm, Vector<double> &last, double &xDiffNorm) const;
  void twinCalcResidualNorm(Vector<double> &x1, double &x1Norm, double &x1InfNorm,
Vector<double> &x2, double &x2Norm, double &x2InfNorm, Vector<double> &last,
double &x1DiffNorm, double &x2DiffNorm) const;

  void print();
};
*/


class StorelessSparseMatrixWithRowColumn{
protected:
  FastDynaArray<StorelessSparseMatrixRow> row;
  FastDynaArray<StorelessSparseMatrixCol> col;
  int non_zeros;
  long rows, cols, elements, calls;
public:
  StorelessSparseMatrixWithRowColumn(long r, long c) {
    rows = r;
    cols = c;
    calls = 0;
    row.setLength(rows);
    col.setLength(cols);
    non_zeros = 0;
    cout << "StorelessSparseMatrixWithRowColumn(" << this << ") constructed with " << rows << " rows and " << cols << " columns..." << endl;
    //    cout << "Length of row=" << row.getLength() << endl;
  }

  int removeState(int target) {

    //    cout << "target for removal=" << target << endl;
    //    row[target].setLength(0);
    //    col[target].setLength(0);

    int entries = row[target].getEntries();
    //    cout << "entries in row " << target << " = " << entries << endl;

    // for each entry in the row, go to the column and get rid of references to that row
    for(int n=0; n<row[target].getEntries(); n++) {
      int column = row[target].getColumn(n);
      col[column].shuffleUp(target);
    }

    // for each entry in the column, go to the row and get rid of references to that column
    for(int m=0; m<col[target].getEntries(); m++) {
      int _row = col[target].getRow(m);
      //      cout << "removing " << target << " from row " << _row << " with " << row[_row].getEntries() << " entries." << endl;
      row[_row].shuffleUp(target);
      //      cout << "after removal we have " << row[_row].getEntries() << endl;
      entries++;     

    }

    //    cout << "entries out = " << entries << endl;
    row[target].remove();
    col[target].remove();

    return entries;

  }

  /*
  void multiply(Vector<double> &x, Vector<double> &result) const {
    register long n,t;
    register double sum;
    register StorelessSparseMatrixRow *_row;

    for (n=0; n<rows; n++) {
      sum = 0; // ### think about extra var here for entries (don't)
      _row = &row(n);
      for (t=0; t<_row->getEntries(); t++) {
	sum +=
	  (_row->getValue(t)) * x(_row->getColumn(t));
    }
    result[n] = sum;
  }
//  cout << "total states: " << total << "\n";
  #ifdef VERBOSE
  cout << "multiply done" << endl;
  #endif
  }
  */

  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getCalls() const { return calls; }
  inline long getElements() const { return elements; }
  inline StorelessSparseMatrixRow &getRow(long index) { return row(index); }
  inline StorelessSparseMatrixCol &getCol(long index) { return col(index); }

  inline void addElement(long _row, long _column, double _entry) {
    row[_row].addElement(_column, _entry);
    col[_column].addElement(_row, _entry);    
    non_zeros++;
  }

  inline int safeAddElement(long _row, long _column, double _entry) {
    double weight = getElement(_row, _column);
    if (weight != 0)
      non_zeros++;
    double new_weight = _entry + weight;    
    //    if(fabs(new_weight) < 1e-16)
    //      cout << "warning new_weight = " << new_weight << endl;
    int result = row[_row].safeAddElement(_column, new_weight);
    col[_column].safeAddElement(_row, new_weight);
    return result;
  }

  inline double getRowElement(long _row, long _index) const {
    return row(_row).getValue(_index); }

  inline double getRowElement(StorelessSparseMatrixRow *_row, long _index) const { return _row->getValue(_index); }

  inline double getColElement(long _col, long _index) const {
    return col(_col).getValue(_index); }

  inline double getColElement(StorelessSparseMatrixCol *_col, long _index) const {    return _col->getValue(_index); }

  inline double getElement(long _row, long _column) const {
    for (long n=0; n<col(_column).getEntries(); n++)
      if (col(_column).getRow(n) == _row) return col(_column).getValue(n);
    return 0.0;     
  }

/*  inline void setElement(long _row, long _column, double _entry) {
    long _index = insert(_entry);
    row[_row].addElement(_column, _index);
    col[_column].addElement(_row, _index);
  }*/

  //  void print();
};

class IntegerStorelessSparseMatrix{
protected:
  FastDynaArray<IntegerStorelessSparseMatrixRow> row;
  FastDynaArray<IntegerStorelessSparseMatrixCol> col;
  int non_zeros;
  long rows, cols, elements, calls;
public:
  IntegerStorelessSparseMatrix(long r, long c) {
    rows = r;
    cols = c;
    calls = 0;
    row.setLength(rows);
    col.setLength(cols);
    non_zeros = 0;
    cout << "IntegerStorelessSparseMatrix(" << this << ") constructed with " << rows << " rows and " << cols << " columns..." << endl;
    //    cout << "Length of row=" << row.getLength() << endl;
  }

  void removeState(int target) {

    //    cout << "target for removal=" << target << endl;
    //    row[target].setLength(0);
    //    col[target].setLength(0);

    row[target].remove();
    col[target].remove();

  }

  /*
  void multiply(Vector<double> &x, Vector<double> &result) const {
    register long n,t;
    register double sum;
    register StorelessSparseMatrixRow *_row;

    for (n=0; n<rows; n++) {
      sum = 0; // ### think about extra var here for entries (don't)
      _row = &row(n);
      for (t=0; t<_row->getEntries(); t++) {
	sum +=
	  (_row->getValue(t)) * x(_row->getColumn(t));
    }
    result[n] = sum;
  }
//  cout << "total states: " << total << "\n";
  #ifdef VERBOSE
  cout << "multiply done" << endl;
  #endif
  }
  */

  inline long getRows() const { return rows; }
  inline long getCols() const { return cols; }
  inline long getCalls() const { return calls; }
  inline long getElements() const { return elements; }
  inline IntegerStorelessSparseMatrixRow &getRow(long index) { return row(index); }
  inline IntegerStorelessSparseMatrixCol &getCol(long index) { return col(index); }

  inline void addElement(long _row, long _column, int _entry) {
    row[_row].addElement(_column, _entry);
    col[_column].addElement(_row, _entry);    
    non_zeros++;
  }

  inline int safeAddElement(long _row, long _column, int _entry) {
    int weight = getElement(_row, _column);
    int new_weight;
    if (weight == -1)
      new_weight = _entry;
    else
      new_weight = _entry - weight;    
    int result = row[_row].safeAddElement(_column, new_weight);
    col[_column].safeAddElement(_row, new_weight);
    return result;
  }

  inline int getRowElement(long _row, long _index) const {
    return row(_row).getValue(_index); }

  inline int getRowElement(IntegerStorelessSparseMatrixRow *_row, long _index) const { return _row->getValue(_index); }

  inline int getColElement(long _col, long _index) const {
    return col(_col).getValue(_index); }

  inline int getColElement(IntegerStorelessSparseMatrixCol *_col, long _index) const {    return _col->getValue(_index); }

  inline int getElement(long _row, long _column) const {
    for (long n=0; n<col(_column).getEntries(); n++)
      if (col(_column).getRow(n) == _row) return col(_column).getValue(n);
    return -1;     
  }

/*  inline void setElement(long _row, long _column, double _entry) {
    long _index = insert(_entry);
    row[_row].addElement(_column, _index);
    col[_column].addElement(_row, _index);
  }*/

  //  void print();
};







