class SteadyStateAnalyser {
  SparseMatrix *Q, *QC;
  SparseMatrixWithRowColumn *QRC;
  DirectMatrix *QD;
  void *core;

  Information *info;
  Timer *timer;

  void terminate(char *fmt, ...);
  long checkVector(Vector<double> &solution, double infNorm);
  void checkViolation(long status);
  void destroy();

  long trans, tangible, genRepInterval, solRepInterval, maxIterations;
  double accuracy, relaxParameter;
  ReportStyle genRepStyle, solRepStyle;
  char *steadyVectorFile;
  FastDynaArray<double> scale;
  FastDynaArray<long> colLength;

  int steps(long);
  MatrixType solver;

  Vector<double> ai_b, ai_d;
  double ai_start;
  void aiPrint(double, double, double, double, double, double);
  void setRandomSeed();
public:
  SteadyStateAnalyser();
  ~SteadyStateAnalyser();
  int readColumnInfo();
  int readStates();
  int readStatesQC(char *filename);
  int readStatesQD(char *filename);
  int readStatesQRC(char *filename);

  void destroyQ();
  void unScale(Vector<double> &);
  void reScale(Vector<double> &);

  void solve();

  void gauss(Vector<double> &);
  void grassmann(Vector<double> &);

  void power(Vector<double> &, long, double);
  void gaussSeidel(Vector<double> &, long, double);
  void sor(Vector<double> &, long, double);
  void sor(double omega, Vector<double> &, long, double);
  void dynamicSOR(Vector<double> &, long, double);

  void cgs(Vector<double> &, long, double);
  void bicgstab(Vector<double> &, long, double);
  void bicgstab2(Vector<double> &, long, double);
  void tfqmr(Vector<double> &, long, double);

  void mycg(Vector<double> &, long, double);
  void update(Vector<double> &x, int k, Vector<double> *h, Vector<double> &s, Vector<double> *v);
  void gpr(double &dx, double &dy, double &cs, double &sn);
  void apr(double &dx, double &dy, double &cs, double &sn);

//  double calcAccuracy(Vector<double> &, Vector<double> &);

  void cgnr(Vector<double> &, long, double);
  void bicg(Vector<double> &, long, double);

  void ai(Vector<double> &, long, double, int);
  void air(Vector<double> &, long, double, int);

  long refine(Vector<double> &, long, double);
  int checkNegative(Vector <double> &);
};

class RowListEntry {
private:
  long column;
  double weight;
  RowListEntry *next;
public:
  RowListEntry(long _column, double _weight) {
    column = _column;
    weight = _weight;
    next = NULL;
  }
  inline long getColumn() { return column; }
  inline double getWeight() { return weight; }
  void setNext(RowListEntry *_next) { next = _next; }
  RowListEntry *getNext() const { return next; }
};

class RowList {
  RowListEntry *front;
public:
  RowList() {
    front = NULL;
  }
  void insert(long _column, double _weight);
  void destroy() {
    RowListEntry *p = front, *q;
    while (p) {
      q = p->getNext();
      delete p;
      p = q;
    }
    front = NULL;
  }
  ~RowList() { destroy(); }
  inline RowListEntry *getFront() { return front; }
};

