////////////////////////////////////////////////////////////////////
// WMCA Markov Chain Analyser        (C) 1995 William Knottenbelt //
////////////////////////////////////////////////////////////////////
//            Data Network Architectures Laboratory               //
//               Department of Computer Science                   //
//                   University of Cape Town                      //
//                       Rondebosch 7700                          //
//                        SOUTH AFRICA                            //
////////////////////////////////////////////////////////////////////
// email: william@cs.uct.ac.za             phone: +27 21 650 2670 //
////////////////////////////////////////////////////////////////////

#define CLOCK

using namespace std;

#include <iostream>
#ifdef CLOCK
#include <ctime>
#else
#ifndef NO_TIMER
#include "time.h"
#include <sys/times.h>
#include <limits>
#endif
#endif
#include "common.hpp"
#include "timer.hpp"


int Timer::start() { 
#ifndef NO_TIMER
#ifdef LOGGING
  cout << "Timer (" << this << "): timer started...\n";
#endif
#ifdef CLOCK
  clock();
  lastTime = 0;
#else
  struct tms buffer;
  times(&buffer);
  lastTime = (double) (buffer.tms_utime + buffer.tms_stime + buffer.tms_cutime + buffer.tms_cstime)/CLK_TCK;
  startTime = lastTime;
#endif
#endif
  return 1;
}

double Timer::read() {
#ifndef NO_TIMER
#ifdef CLOCK
  double timeNow = (double) clock()/CLOCKS_PER_SEC;
#else
  struct tms buffer;
  times(&buffer);
  double timeNow = (double) (buffer.tms_utime + buffer.tms_stime + buffer.tms_cutime + buffer.tms_cstime)/CLK_TCK; 
#endif
  double elapsed = (timeNow - lastTime) > 0 ? (timeNow - lastTime) : 0;
#ifdef LOGGING
  cout << "Timer (" << this << "): reading timer (" << elapsed << " seconds)\n";
#endif
  lastTime = timeNow;
  return elapsed;
#else
  return 0;
#endif
}

double Timer::observe()
{
#ifndef NO_TIMER
#ifdef CLOCK
  return (double) clock()/CLOCKS_PER_SEC;
#else
  struct tms buffer;
  times(&buffer);
  double timeNow = (double) (buffer.tms_utime + buffer.tms_stime + buffer.tms_cutime + buffer.tms_cstime)/CLK_TCK;
  return timeNow - startTime;
#endif
#else
  return 0;
#endif //;
}

