enum Token {

  // Keywords

  T_MODEL_DESCRIPTION,     T_STATE_VECTOR,          T_TYPE,
  T_HEADER,                T_TRANSITION,
  T_CONSTRUCTOR,           T_DESTRUCTOR,
  T_INITIAL,               T_CONDITION,             T_ACTION,
  T_RATE,                  T_WEIGHT,
  T_PRIORITY,              T_GENERATION_CONTROL,    T_MAX_STATES,
  T_MAX_CPU_TIME,          T_SAFE_HASH,              T_REPORT_STYLE,         
  T_REPORT_INTERVAL,       T_CONSTANT,              T_HELP_VALUE,
  T_PRIMARY,               T_SECONDARY,             T_INVARIANT,
  T_SOLUTION_CONTROL,      T_METHOD,                T_ACCURACY,
  T_MAX_ITERATIONS,        T_RELAX_PARAMETER,       
  T_START_VECTOR,          T_PERFORMANCE,           T_STATE_MEASURE,
  T_COUNT_MEASURE,         T_ESTIMATOR,             T_EXPRESSION,
  T_PRECONDITION,          T_POSTCONDITION,         T_OUTPUT,
  T_STATELIST,             T_STEADY_STATE_VECTOR,   T_TRANSITION_MATRIX,
  T_PERFORMANCE_RESULTS,   T_PASSAGE,               T_TARGET_CONDITION,
  T_SOURCE_CONDITION,
  T_SOJOURNTIME_LT,        T_START,                 T_STOP,
  T_STEP,                  T_SOJOURNTIME,           T_EXCLUDE_CONDITION,
  T_TRANSIENT,            

  // Parameters

  T_GAUSS,                 T_LU,                    T_GRASSMANN,
  T_POWER,
  T_GAUSS_SEIDEL,          T_SOR,                   T_CGNR,
  T_CGS,
  T_BICG,                  T_BICGSTAB,              T_BICGSTAB2,
  T_MYCG,
  T_AI,                    T_AIR,
  T_AUTOMATIC,             T_ON, 
  T_OFF,                   T_FULL,                  T_BRIEF,
  T_NONE,                  T_DYNAMIC,
  T_DIAG,                  T_ILU,                   T_MEAN,
  T_VARIANCE,              T_STDDEV,                T_DISTRIBUTION,
  T_ALL,    

  // Identifiers
  T_IDENTIFIER,

  // punctuation
  T_OPEN_BRACE, T_CLOSE_BRACE, T_COMMA, T_SEMICOLON, T_PERCENT, T_EQUAL, T_OPERATOR,
  T_DEREFERENCE,

  // other
  T_EOF, T_UNDEFINED
};


typedef struct KeywordEntry {
  Token token;
  char *keyword;
} KeywordEntry;

class CodeBlock : public FastDynaArray<char> {
public:
  CodeBlock() {
  } // ### FIX
  ~CodeBlock() {}
  friend ostream &operator <<(ostream &out, const CodeBlock &source);
};

class Tokeniser {
protected:
  ifstream  source;
  char      ch, identifier[MAX_LENGTH], sourceFile[MAX_LENGTH];
  long      lineNumber;
  int       errors, warnings, suppressComment;
  Token     current;

  int       nextChar();
  void format(char *message);
public:
  Tokeniser();

  int       openSource(char *filename);
  Token     getToken();
  Token     getNextToken();
  char *    getTokenText(Token);
  char *    getIdentifier() { return identifier; }
  int       expect(Token expected);
  int       checkInteger();
  Token     scanIdentifier();
  Token     scanReal();
  int       scanCodeBlock(CodeBlock &, int = 0);
  int       error(const char *fmt, ...);
  int       warning(const char *fmt, ...);
  int       unexpected(Token token, char * context);
  int       find(Token token);
  inline long getLineNumber() { return lineNumber; }
  inline int getErrors() { return errors; }
  inline int getWarnings() { return warnings; }
  inline int whitespace(char ch) {
    if (ch=='\n' || ch == '\t' || ch == ' ' || ch == 10) return 1;
    return 0;
  }
  inline int report() {
    if (errors > 0 || warnings > 0)
      printf("\n%3d error(s) and %3d warning(s) during parsing of '%s'.\n", errors, warnings, sourceFile);
    return 1;
  }
};

